#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import os
from lxml import etree  # pyright: ignore[reportAttributeAccessIssue]
import urllib.request as ur
from Bio.SeqIO.FastaIO import SimpleFastaParser

from ..data import check_input
from .to_proxl_util import __local_schema
from .util import __get_filename
from .to_alphalink2 import __protein_supported_by_crosslink
from ..transform.util import get_available_keys
from ..transform.util import modifications_to_str as mts
from ..constants import MODIFICATIONS

from typing import Optional
from typing import Callable
from typing import List
from typing import Dict
from typing import Tuple
from typing import Any

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

SCHEMA_URL = "https://github.com/yeastrc/proxl-import-api/raw/refs/heads/master/xsd/proxl-xml.xsd"


def __build_header(
    fasta_filename: str,
    search_engine: str,
    version: str,
    score: Literal["higher_better", "lower_better"],
    crosslinker: str,
    crosslinker_mass: float,
) -> List[str]:
    r"""Builds the header section of the ProXL XML.

    Parameters
    ----------
    fasta_filename : str
        The name/path of the fasta file for reading protein sequences.
    search_engine : str
        Name of the used crosslink search engine.
    version : str
        Version identifier of the used crosslink search engine.
    score : str, one of "higher_better" or "lower_better"
        If a higher score is considered better, or a lower score is considered better.
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    crosslinker_mass : float
        Monoisotopic delta mass of the crosslink modification.

    Returns
    -------
    list of str
        A list of lines of the header section of the ProXL XML.

    Notes
    -----
    This function should not be called directly, it is called from ``to_proxl()``.
    """
    filter_direction = "above" if score == "higher_better" else "below"
    lines = [
        r"""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>""",
        f'<proxl_input fasta_filename="{fasta_filename}">',
        r"""<search_program_info>""",
        r"""<search_programs>""",
        f'<search_program name="{search_engine}" display_name="{search_engine}" version="{version}">',
        r"""<psm_annotation_types>""",
        r"""<filterable_psm_annotation_types>""",
        f'<filterable_psm_annotation_type name="score" description="Score of the search engine" filter_direction="{filter_direction}" default_filter="false"/>',
        r"""</filterable_psm_annotation_types>""",
        r"""<descriptive_psm_annotation_types>""",
        r"""<descriptive_psm_annotation_type name="spectrum filename" description="Name of the MS file"/>""",
        r"""<descriptive_psm_annotation_type name="scan number" description="Scan number"/>""",
        r"""</descriptive_psm_annotation_types>""",
        r"""</psm_annotation_types>""",
        r"""</search_program>""",
        r"""</search_programs>""",
        r"""<default_visible_annotations>""",
        r"""<visible_psm_annotations>""",
        f'<search_annotation search_program="{search_engine}" annotation_name="score"/>',
        f'<search_annotation search_program="{search_engine}" annotation_name="spectrum filename"/>',
        f'<search_annotation search_program="{search_engine}" annotation_name="scan number"/>',
        r"""</visible_psm_annotations>""",
        r"""</default_visible_annotations>""",
        r"""</search_program_info>""",
        r"""<linkers>""",
        f'<linker name="{crosslinker}">',
        r"""<crosslink_masses>""",
        f'<crosslink_mass mass="{crosslinker_mass}"/>',
        r"""</crosslink_masses>""",
        r"""</linker>""",
        r"""</linkers>""",
    ]
    return lines


def __get_reported_peptide_string(csm: Dict[str, Any]) -> str:
    r"""Creates a unique 'reported_peptide_string' for a crosslink-spectrum-match.

    Parameters
    ----------
    csm : dict of str, any
        A crosslink-spectrum-match.

    Returns
    -------
    str
        The unique 'reported_peptide_string' of the crosslink-spectrum-match.

    Notes
    -----
    This function should not be called directly, it is called from ``to_proxl()``.
    """
    return (
        f"{csm['alpha_peptide']}({csm['alpha_peptide_crosslink_position']})-{csm['beta_peptide']}({csm['beta_peptide_crosslink_position']})"
        f"_({mts(csm['alpha_modifications'])})-({mts(csm['beta_modifications'])})"
    )


def __get_reported_peptides(
    csms: List[Dict[str, Any]],
) -> Dict[str, List[Dict[str, Any]]]:
    r"""Groups crosslink-spectrum-matches by their unique 'reported_peptide_string'.

    Parameters
    ----------
    csms : list of dict of str, any
        A list of crosslink-spectrum-matches.

    Returns
    -------
    dict of str, list of dict of str, any
        A dictionary that maps unique 'reported_peptide_string' keys to lists of associated
        crosslink-spectrum-matches as values.

    Notes
    -----
    This function should not be called directly, it is called from ``to_proxl()``.
    """
    reported_peptides = dict()
    for csm in csms:
        reported_peptide_string = __get_reported_peptide_string(csm)
        if reported_peptide_string in reported_peptides:
            reported_peptides[reported_peptide_string].append(csm)
        else:
            reported_peptides[reported_peptide_string] = [csm]
    return reported_peptides


def __build_psm(
    csm: Dict[str, Any], crosslinker_mass: float, search_engine: str
) -> List[str]:
    r"""Builds the 'psm' section of the ProXL XML.

    Parameters
    ----------
    csms : dict of str, any
        A crosslink-spectrum-match.
    crosslinker_mass : float
        Monoisotopic delta mass of the crosslink modification.
    search_engine : str
        Name of the used crosslink search engine.

    Returns
    -------
    list of str
        A list of lines of one 'psm' section of the ProXL XML.

    Notes
    -----
    This function should not be called directly, it is called from ``to_proxl()``.
    """
    scan_file_name = os.path.splitext(csm["spectrum_file"])[0] + ".mzML"
    lines = [
        f'<psm scan_file_name="{scan_file_name}" scan_number="{csm["scan_nr"]}" precursor_charge="{csm["charge"]}" linker_mass="{crosslinker_mass}">',
        r"""<filterable_psm_annotations>""",
        f'<filterable_psm_annotation search_program="{search_engine}" annotation_name="score" value="{csm["score"]}"/>',
        r"""</filterable_psm_annotations>""",
        r"""<descriptive_psm_annotations>""",
        f'<descriptive_psm_annotation search_program="{search_engine}" annotation_name="spectrum filename" value="{scan_file_name}"/>',
        f'<descriptive_psm_annotation search_program="{search_engine}" annotation_name="scan number" value="{csm["scan_nr"]}"/>',
        r"""</descriptive_psm_annotations>""",
        r"""</psm>""",
    ]
    return lines


def __build_modifications(csm: Dict[str, Any]) -> Tuple[List[str], List[str]]:
    r"""Builds the 'modifications' section of the ProXL XML.

    Parameters
    ----------
    csm : dict of str, any
        A crosslink-spectrum-match.

    Returns
    -------
    tuple of list of str, list of str
        A list of lines of one 'modifications' section of the ProXL XML for each crosslinked
        peptide.

    Notes
    -----
    This function should not be called directly, it is called from ``to_proxl()``.
    """
    modifications_a = list()
    if csm["alpha_modifications"] is not None:
        modifications_a.append(r"""<modifications>""")
        for pos, modification in csm["alpha_modifications"].items():
            if pos == 0:
                pos = 1
            if pos > len(csm["alpha_peptide"]):
                pos = len(csm["alpha_peptide"])
            if pos != csm["alpha_peptide_crosslink_position"]:
                modifications_a.append(
                    f'<modification mass="{modification[1]}" position="{pos}" isMonolink="false"/>'
                )
        modifications_a.append(r"""</modifications>""")
    if len(modifications_a) <= 2:
        modifications_a.clear()
    modifications_b = list()
    if csm["beta_modifications"] is not None:
        modifications_b.append(r"""<modifications>""")
        for pos, modification in csm["beta_modifications"].items():
            if pos == 0:
                pos = 1
            if pos > len(csm["beta_peptide"]):
                pos = len(csm["beta_peptide"])
            if pos != csm["beta_peptide_crosslink_position"]:
                modifications_b.append(
                    f'<modification mass="{modification[1]}" position="{pos}" isMonolink="false"/>'
                )
        modifications_b.append(r"""</modifications>""")
    if len(modifications_b) <= 2:
        modifications_b.clear()
    return (modifications_a, modifications_b)


def __build_reported_peptides(
    reported_peptides: Dict[str, List[Dict[str, Any]]],
    crosslinker_mass: float,
    search_engine: str,
) -> List[str]:
    r"""Builds the 'reported_peptides' section of the ProXL XML.

    Parameters
    ----------
    reported_peptides : dict of str, list of dict of str, any
        A dictionary that maps unique 'reported_peptide_string' keys to lists of associated
        crosslink-spectrum-matches as values.
    crosslinker_mass : float
        Monoisotopic delta mass of the crosslink modification.
    search_engine : str
        Name of the used crosslink search engine.

    Returns
    -------
    list of str
        A list of lines of the 'reported_peptides' section of the ProXL XML.

    Notes
    -----
    This function should not be called directly, it is called from ``to_proxl()``.
    """
    lines = [r"""<reported_peptides>"""]
    for reported_peptide in reported_peptides:
        example_csm = reported_peptides[reported_peptide][0]
        modifications = __build_modifications(example_csm)
        lines += [
            f'<reported_peptide reported_peptide_string="{reported_peptide}" type="crosslink">',
            r"""<peptides>""",
            f'<peptide sequence="{example_csm["alpha_peptide"]}">',
        ]
        lines += modifications[0]
        lines += [
            r"""<linked_positions>""",
            f'<linked_position position="{example_csm["alpha_peptide_crosslink_position"]}"/>',
            r"""</linked_positions>""",
            r"""</peptide>""",
            f'<peptide sequence="{example_csm["beta_peptide"]}">',
        ]
        lines += modifications[1]
        lines += [
            r"""<linked_positions>""",
            f'<linked_position position="{example_csm["beta_peptide_crosslink_position"]}"/>',
            r"""</linked_positions>""",
        ]
        lines.append(r"""</peptide>""")
        lines.append(r"""</peptides>""")
        lines.append(r"""<psms>""")
        for csm in reported_peptides[reported_peptide]:
            lines += __build_psm(csm, crosslinker_mass, search_engine)
        lines.append(r"""</psms>""")
        lines.append(r"""</reported_peptide>""")
    lines.append(r"""</reported_peptides>""")
    return lines


def __build_matched_proteins(
    csms: List[Dict[str, Any]],
    fasta_filename: str,
    title_to_accession: Optional[Callable[[str], str]],
) -> List[str]:
    r"""Builds the 'matched_proteins' section of the ProXL XML.

    Parameters
    ----------
    csms : list of dict of str, any
        A list of crosslink-spectrum-matches.
    fasta_filename : str
        The name/path of the fasta file for reading protein sequences.
    title_to_accession : callable, or None
        A function that parses the protein accession from the fasta title/header. If None (default)
        the full fasta headers are used. An example function would be ``transform.fasta_title_to_accession()``.

    Returns
    -------
    list of str
        A list of lines of the 'matched_proteins' section of the ProXL XML.

    Notes
    -----
    This function should not be called directly, it is called from ``to_proxl()``.
    """
    lines = [r"""<matched_proteins>"""]
    fasta_items = list()
    with open(fasta_filename, "r", encoding="utf-8") as f:
        for item in SimpleFastaParser(f):
            fasta_items.append(item)
    for item in fasta_items:
        header = item[0].replace('"', "")
        name = title_to_accession(header) if title_to_accession is not None else header
        sequence = item[1]
        if __protein_supported_by_crosslink(sequence, csms):
            lines.append(f'<protein sequence="{sequence}">')
            lines.append(f'<protein_annotation name="{name}"/>')
            lines.append(r"""</protein>""")
    lines.append(r"""</matched_proteins>""")
    return lines


def __validate_schema(
    xml_str: str, schema_validation: Literal["online", "offline"]
) -> bool:
    r"""Validates a ProXL XML string against the ProXL XML schema.

    Parameters
    ----------
    xml_str : str
        The ProXL XML string to validate.
    schema_validation : str, one of "online" or "offline"
        If XML schema validation should use the most recent online schema or the locally stored but
        possibly outdated offline schema.

    Returns
    -------
    bool
        True if the ProXL XML string validates successfully against the schema, otherwise False.

    Notes
    -----
    This function should not be called directly, it is called from ``to_proxl()``.
    """
    proxl_schema = __local_schema.encode("utf-8")
    if schema_validation == "online":
        proxl_schema = ur.urlopen(SCHEMA_URL).read()
    parser = etree.XMLParser(encoding="utf-8")
    schema_doc = etree.fromstring(proxl_schema, parser=parser)
    schema = etree.XMLSchema(schema_doc)
    xml_doc = etree.fromstring(xml_str.encode("utf-8"), parser=parser)
    return schema.validate(xml_doc)


def to_proxl(
    csms: List[Dict[str, Any]],
    fasta_filename: str,
    search_engine: str,
    search_engine_version: str,
    score: Literal["higher_better", "lower_better"],
    crosslinker: str,
    crosslinker_mass: Optional[float] = None,
    modifications: Dict[str, float] = MODIFICATIONS,
    fasta_filename_override: Optional[str] = None,
    fasta_title_to_accession: Optional[Callable[[str], str]] = None,
    filename: Optional[str] = None,
    schema_validation: Literal["online", "offline"] = "online",
) -> str:
    r"""Exports a list of crosslink-spectrum-matches to ProXL format.

    Exports a list of crosslink-spectrum-matches to ProXL format. The tool ProXL is accessible
    via the link `yeastrc.org/proxl_public <https://www.yeastrc.org/proxl_public/>`_.
    Requires that ``charge`` and ``score`` fields are set for all crosslink-spectrum-matches.

    Parameters
    ----------
    csms : list of dict of str, any
        A list of crosslink-spectrum-matches.
    fasta_filename : str
        The name/path of the fasta file for reading protein sequences.
    search_engine : str
        Name of the used crosslink search engine.
    search_engine_version : str
        Version identifier of the used crosslink search engine.
    score : str, one of "higher_better" or "lower_better"
        If a higher score is considered better, or a lower score is considered better.
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    crosslinker_mass : float, or None, default = None
        Monoisotopic delta mass of the crosslink modification. If the crosslinker is
        defined in parameter "modifications" this can be omitted.
    modifications: dict of str, float, default = ``constants.MODIFICATIONS``
        Mapping of modification names to modification masses. By default uses ``constants.MODIFICATIONS``.
    fasta_filename_override : str, or None, default = None
        Name that should be used in the ProXL XML for the fasta file. If None (default)
        uses the filename of parameter 'fasta_filename' (preceding directories are pruned).
    fasta_title_to_accession : callable, or None, default = None
        A function that parses the protein accession from the fasta title/header. If None (default)
        the full fasta headers are used. An example function would be ``transform.fasta_title_to_accession()``.
    filename : str, or None, default = None
        If not None, the exported data will be written to a file with the specified filename.
    schema_validation : str, one of "online" or "offline", default = "online"
        If XML schema validation should use the most recent online schema or the locally stored but
        possibly outdated offline schema.

    Returns
    -------
    str
        The ProXL XML as a string.

    Raises
    ------
    TypeError
        If a wrong data type is provided.
    TypeError
        If 'csms' parameter contains elements of mixed data type.
    TypeError
        If parameter score is not one of 'higher_better' or 'lower_better'.
    TypeError
        If parameter schema_validation is not one of 'online' or 'offline'.
    ValueError
        If the provided 'csms' parameter contains no elements.
    KeyError
        If the specified crosslinker could not be found/mapped.
    RuntimeError
        If not all of the required information is present in the input data.
    RuntimeError
        If the created ProXL XML file fails validation against the ProXL XML schema.

    Examples
    --------
    >>> from pyXLMS.pipelines import pipeline
    >>> from pyXLMS.exporter import to_proxl
    >>> pr = pipeline(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> xml = to_proxl(
    ...     pr["crosslink-spectrum-matches"],
    ...     fasta_filename="data/_fasta/Cas9_plus10.fasta",
    ...     search_engine="MS Annika",
    ...     search_engine_version="3.0.1",
    ...     score="higher_better",
    ...     crosslinker="DSS",
    ...     filename="DSS_Cas9_ProXL.xml",
    ... )
    """
    _ok = check_input(csms, "csms", list, dict)
    _ok = check_input(fasta_filename, "fasta_filename", str)
    _ok = check_input(search_engine, "search_engine", str)
    _ok = check_input(search_engine_version, "search_engine_version", str)
    _ok = check_input(score, "score", str)
    _ok = check_input(crosslinker, "crosslinker", str)
    _ok = (
        check_input(crosslinker_mass, "crosslink_mass", float)
        if crosslinker_mass is not None
        else True
    )
    _ok = check_input(modifications, "modifications", dict, float)
    _ok = (
        check_input(fasta_filename_override, "fasta_filename_override", str)
        if fasta_filename_override is not None
        else True
    )
    _ok = (
        check_input(fasta_title_to_accession, "title_to_accession", Callable)
        if fasta_title_to_accession is not None
        else True
    )
    _ok = check_input(filename, "filename", str) if filename is not None else True
    _ok = check_input(schema_validation, "schema_validation", str)
    if score not in ["higher_better", "lower_better"]:
        raise TypeError(
            "Parameter 'score' has to be one of 'higher_better' or 'lower_better'!"
        )
    if schema_validation not in ["online", "offline"]:
        raise TypeError(
            "Parameter 'schema_validation' has to be one of 'online' or 'offline'!"
        )
    if crosslinker_mass is None:
        if crosslinker not in modifications:
            raise KeyError(
                "Cannot infer crosslinker mass because crosslinker is not defined in "
                "parameter 'modifications'. Please specify crosslinker mass manually!"
            )
        else:
            crosslinker_mass = modifications[crosslinker]
    if len(csms) == 0:
        raise ValueError("Provided crosslink-spectrum-matches contain no elements!")
    if "data_type" not in csms[0] or csms[0]["data_type"] != "crosslink-spectrum-match":
        raise TypeError(
            "Unsupported data type for input csms! Parameter csms has to be a list of crosslink-spectrum-matches!"
        )
    available_keys = get_available_keys(csms)
    if not available_keys["score"] or not available_keys["charge"]:
        raise RuntimeError(
            "Can't export to ProXL because not all necessary information is available!"
        )
    fasta_name = (
        os.path.basename(fasta_filename)
        if fasta_filename_override is None
        else fasta_filename_override
    )
    reported_peptides = __get_reported_peptides(csms)
    lines = (
        __build_header(
            fasta_name,
            search_engine,
            search_engine_version,
            score,
            crosslinker,
            crosslinker_mass,
        )
        + __build_reported_peptides(reported_peptides, crosslinker_mass, search_engine)
        + __build_matched_proteins(csms, fasta_filename, fasta_title_to_accession)
        + [r"""</proxl_input>"""]
    )
    xml_str = "\n".join(lines)
    if __validate_schema(xml_str, schema_validation):
        print(
            f"Successfully created ProXL XML and validated it against {schema_validation} XML schema!"
        )
    else:
        raise RuntimeError(
            f"Created ProXL XML but validation against {schema_validation} XML schema failed!"
        )
    if filename is not None:
        with open(__get_filename(filename, "xml"), "w", encoding="utf-8") as f:
            f.write(xml_str)
            f.close()
    return xml_str



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import pandas as pd

from ..data import check_input
from ..transform.util import get_available_keys
from .to_xinet import to_xinet
from .util import __get_filename

from typing import Optional
from typing import Dict
from typing import Any
from typing import List
from typing import Callable


def __xls_to_xiview_minimal(
    xls: List[Dict[str, Any]],
    filename: Optional[str],
) -> pd.DataFrame:
    r"""Exports crosslinks to xiVIEW format.

    Parameters
    ----------
    xls : list of dict of str, any
        A list of crosslinks.
    filename : str, or None
        If not None, the data will be written to a file with the specified filename.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame in xiVIEW format.

    Notes
    -----
    This function should not be called directly, it is called from ``to_xiview()``.
    """
    abspos1 = list()
    abspos2 = list()
    protein1 = list()
    protein2 = list()
    decoy1 = list()
    decoy2 = list()
    score = list()
    has_decoys = True
    has_scores = True

    for xl in xls:
        abspos1.append(
            ";".join([str(pos) for pos in xl["alpha_proteins_crosslink_positions"]])
        )
        abspos2.append(
            ";".join([str(pos) for pos in xl["beta_proteins_crosslink_positions"]])
        )
        protein1.append(";".join(xl["alpha_proteins"]))
        protein2.append(";".join(xl["beta_proteins"]))
        if has_decoys:
            if xl["alpha_decoy"] is not None and xl["beta_decoy"] is not None:
                if xl["alpha_decoy"]:
                    decoy1.append(";".join(["TRUE" for p in xl["alpha_proteins"]]))
                else:
                    decoy1.append(";".join(["FALSE" for p in xl["alpha_proteins"]]))
                if xl["beta_decoy"]:
                    decoy2.append(";".join(["TRUE" for p in xl["beta_proteins"]]))
                else:
                    decoy2.append(";".join(["FALSE" for p in xl["beta_proteins"]]))
            else:
                has_decoys = False
        if has_scores:
            if xl["score"] is not None:
                score.append(xl["score"])
            else:
                has_scores = False

    xiview_df = pd.DataFrame(
        {
            "AbsPos1": abspos1,
            "AbsPos2": abspos2,
            "Protein1": protein1,
            "Protein2": protein2,
        }
    )

    if has_decoys:
        xiview_df["Decoy1"] = decoy1
        xiview_df["Decoy2"] = decoy2
    if has_scores:
        xiview_df["Score"] = score

    if filename is not None:
        xiview_df.to_csv(__get_filename(filename, "csv"), index=False)
    return xiview_df


def __get_PeakListFileName(filename: str) -> str:
    ## replaces file extension with "mzML"
    return ".".join(filename.split(".")[:-1]) + ".mzML"


def __csms_to_xiview_with_peaks(
    csms: List[Dict[str, Any]],
    crosslinker_mass: float,
    get_PeakListFileName: Optional[Callable[[str], str]] = None,
    filename: Optional[str] = None,
    unsafe: bool = False,
) -> pd.DataFrame:
    ## experimental export to CSV with peaks list
    ## see -> https://xiview.org/csv-formats.php
    ##
    ## params:
    ## csms -> list of crosslink-spectrum-matches
    ## crosslinker_mass -> mono-isotopic delta modification mass of the used crosslinker
    ## get_PeakListFileName -> function that maps csm["spectrum_file"] to PeakListFileName, by default: __get_PeakListFileName
    ## filename -> if given, data frame is written to file
    ## unsafe -> needs to be True in order to run, otherwise throws NotImplementedError
    ##
    ## warning!
    ## this function is untested on real data!
    ## should also only be used with mzML files, because of ScanId definition (see https://xiview.org/csv-formats.php)!
    ##
    ## required crosslink-spectrum-match attributes:
    ## - alpha_proteins_crosslink_positions
    ## - beta_proteins_crosslink_positions
    ## - alpha_proteins
    ## - beta_proteins
    ## - charge
    if not unsafe:
        raise NotImplementedError()
    if get_PeakListFileName is None:
        get_PeakListFileName = __get_PeakListFileName

    pepseq1 = list()
    pepseq2 = list()
    peppos1 = list()
    peppos2 = list()
    linkpos1 = list()
    linkpos2 = list()
    protein1 = list()
    protein2 = list()
    charge = list()
    crosslinkermodmass = list()
    scanid = list()
    peaklistfilename = list()
    decoy1 = list()
    decoy2 = list()
    score = list()
    has_decoys = True
    has_scores = True

    for csm in csms:
        pos1 = csm["alpha_peptide_crosslink_position"]
        pos2 = csm["beta_peptide_crosslink_position"]
        pepseq1.append(csm["alpha_peptide"])
        pepseq2.append(csm["beta_peptide"])
        peppos1.append(
            ";".join(
                [
                    str(pos - pos1 + 1)
                    for pos in csm["alpha_proteins_crosslink_positions"]
                ]
            )
        )
        peppos2.append(
            ";".join(
                [
                    str(pos - pos2 + 1)
                    for pos in csm["beta_proteins_crosslink_positions"]
                ]
            )
        )
        linkpos1.append(pos1)
        linkpos2.append(pos2)
        protein1.append(";".join(csm["alpha_proteins"]))
        protein2.append(";".join(csm["beta_proteins"]))
        charge.append(csm["charge"])
        crosslinkermodmass.append(crosslinker_mass)
        scanid.append(csm["scan_nr"])
        peaklistfilename.append(get_PeakListFileName(csm["spectrum_file"]))
        if has_decoys:
            if csm["alpha_decoy"] is not None and csm["beta_decoy"] is not None:
                if csm["alpha_decoy"]:
                    decoy1.append(";".join(["TRUE" for p in csm["alpha_proteins"]]))
                else:
                    decoy1.append(";".join(["FALSE" for p in csm["alpha_proteins"]]))
                if csm["beta_decoy"]:
                    decoy2.append(";".join(["TRUE" for p in csm["beta_proteins"]]))
                else:
                    decoy2.append(";".join(["FALSE" for p in csm["beta_proteins"]]))
            else:
                has_decoys = False
        if has_scores:
            if csm["score"] is not None:
                score.append(csm["score"])
            else:
                has_scores = False

    xiview_df = pd.DataFrame(
        {
            "PepSeq1": pepseq1,
            "PepSeq2": pepseq2,
            "PepPos1": peppos1,
            "PepPos2": peppos2,
            "LinkPos1": linkpos1,
            "LinkPos2": linkpos2,
            "Protein1": protein1,
            "Protein2": protein2,
            "Charge": charge,
            "CrossLinkerModMass": crosslinkermodmass,
            "ScanId": scanid,
            "PeakListFileName": peaklistfilename,
        }
    )
    if has_decoys:
        xiview_df["Decoy 1"] = decoy1
        xiview_df["Decoy 2"] = decoy2
    if has_scores:
        xiview_df["Score"] = score

    if filename is not None:
        xiview_df.to_csv(__get_filename(filename, "csv"), index=False)
    return xiview_df


def to_xiview(
    crosslinks: List[Dict[str, Any]],
    filename: Optional[str],
    minimal: bool = True,
) -> pd.DataFrame:
    r"""Exports a list of crosslinks to xiVIEW format.

    Exports a list of crosslinks to xiVIEW format. The tool xiVIEW is accessible
    via the link
    `xiview.org/ <https://xiview.org/>`_.
    Requires that ``alpha_proteins``, ``beta_proteins``, ``alpha_proteins_crosslink_positions`` and
    ``beta_proteins_crosslink_positions`` fields are set for all crosslinks.

    Parameters
    ----------
    crosslinks : list of dict of str, any
        A list of crosslinks.
    filename : str, or None
        If not None, the exported data will be written to a file with the specified filename.
    minimal : bool, default = True
        Which xiVIEW format to return, if ``minimal = True`` the minimal xiVIEW format is returned. Otherwise
        the "CSV without peak lists" format is returned (internally this just calls ``exporter.to_xinet()``).
        For more information on the xiVIEW formats please refer to the `xiVIEW specification <https://xiview.org/csv-formats.php>`_.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame containing crosslinks in xiVIEW format.

    Raises
    ------
    TypeError
        If a wrong data type is provided.
    TypeError
        If 'crosslinks' parameter contains elements of mixed data type.
    ValueError
        If the provided 'crosslinks' parameter contains no elements.
    RuntimeError
        If not all of the required information is present in the input data.

    Notes
    -----
    The optional ``Score`` column in the xiVIEW table will only be available if all crosslinks have assigned scores,
    the optional ``Decoy*`` columns will only be available if all crosslinks have assigned target and decoy labels.

    Examples
    --------
    >>> from pyXLMS.exporter import to_xiview
    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import targets_only
    >>> from pyXLMS.transform import filter_proteins
    >>> pr = read(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> crosslinks = targets_only(pr)["crosslinks"]
    >>> cas9 = filter_proteins(crosslinks, proteins=["Cas9"])["Both"]
    >>> to_xiview(cas9, filename="crosslinks_xiVIEW.csv")
        AbsPos1 AbsPos2 Protein1 Protein2 Decoy1 Decoy2   Score
    0       779     779     Cas9     Cas9  FALSE  FALSE  119.83
    1       866     866     Cas9     Cas9  FALSE  FALSE  114.43
    2       677     677     Cas9     Cas9  FALSE  FALSE  200.98
    3       677      48     Cas9     Cas9  FALSE  FALSE   94.47
    4        34      34     Cas9     Cas9  FALSE  FALSE  110.48
    ..      ...     ...      ...      ...    ...    ...     ...
    248     396     396     Cas9     Cas9  FALSE  FALSE  305.63
    249     688     952     Cas9     Cas9  FALSE  FALSE  110.46
    250     793    1180     Cas9     Cas9  FALSE  FALSE  288.36
    251     575     688     Cas9     Cas9  FALSE  FALSE  376.15
    252    1180    1180     Cas9     Cas9  FALSE  FALSE  437.10
    [253 rows x 7 columns]

    >>> from pyXLMS.exporter import to_xiview
    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import targets_only
    >>> from pyXLMS.transform import filter_proteins
    >>> pr = read(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> crosslinks = targets_only(pr)["crosslinks"]
    >>> cas9 = filter_proteins(crosslinks, proteins=["Cas9"])["Both"]
    >>> df = to_xiview(cas9, filename=None)

    >>> from pyXLMS.exporter import to_xiview
    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import targets_only
    >>> from pyXLMS.transform import filter_proteins
    >>> pr = read(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> crosslinks = targets_only(pr)["crosslinks"]
    >>> cas9 = filter_proteins(crosslinks, proteins=["Cas9"])["Both"]
    >>> to_xiview(cas9, filename="crosslinks_xiVIEW.csv", minimal=False)
        Protein1 PepPos1           PepSeq1  LinkPos1 Protein2 PepPos2         PepSeq2  LinkPos2   Score   Id
    0       Cas9     777            GQKNSR         3     Cas9     777          GQKNSR         3  119.83    1
    1       Cas9     864             SDKNR         3     Cas9     864           SDKNR         3  114.43    2
    2       Cas9     676            DKQSGK         2     Cas9     676          DKQSGK         2  200.98    3
    3       Cas9     676            DKQSGK         2     Cas9      45           HSIKK         4   94.47    4
    4       Cas9      31             VPSKK         4     Cas9      31           VPSKK         4  110.48    5
    ..       ...     ...               ...       ...      ...     ...             ...       ...     ...  ...
    248     Cas9     387     MDGTEELLVKLNR        10     Cas9     387   MDGTEELLVKLNR        10  305.63  249
    249     Cas9     682    TILDFLKSDGFANR         7     Cas9     947       YDENDKLIR         6  110.46  250
    250     Cas9     788    IEEGIKELGSQILK         6     Cas9    1176  SSFEKNPIDFLEAK         5  288.36  251
    251     Cas9     575  KIECFDSVEISGVEDR         1     Cas9     682  TILDFLKSDGFANR         7  376.15  252
    252     Cas9    1176    SSFEKNPIDFLEAK         5     Cas9    1176  SSFEKNPIDFLEAK         5  437.10  253
    [253 rows x 10 columns]
    """
    _ok = check_input(crosslinks, "crosslinks", list, dict)
    _ok = check_input(filename, "filename", str) if filename is not None else True
    if len(crosslinks) == 0:
        raise ValueError("Provided crosslinks contain no elements!")
    if "data_type" not in crosslinks[0] or crosslinks[0]["data_type"] != "crosslink":
        raise TypeError(
            "Unsupported data type for input crosslinks! Parameter crosslinks has to be a list of crosslinks!"
        )
    available_keys = get_available_keys(crosslinks)
    if (
        not available_keys["alpha_proteins"]
        or not available_keys["beta_proteins"]
        or not available_keys["alpha_proteins_crosslink_positions"]
        or not available_keys["beta_proteins_crosslink_positions"]
    ):
        raise RuntimeError(
            "Can't export to xiVIEW because not all necessary information is available!"
        )
    if minimal:
        return __xls_to_xiview_minimal(crosslinks, filename)
    return to_xinet(crosslinks, filename)



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import pandas as pd

from ..data import check_input
from ..transform.util import assert_data_type_same
from .util import __get_filename

from typing import Optional
from typing import Dict
from typing import Any
from typing import List

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def get_msannika_crosslink_sequence(peptide: str, crosslink_position: int) -> str:
    r"""Returns the crosslinked peptide sequence in MS Annika format.

    Returns the crosslinked peptide sequence in MS Annika format, which is the peptide amino
    acid sequence with the crosslinked residue in square brackets (see examples).

    Parameters
    ----------
    peptide : str
        The (unmodified) amino acid sequence of the peptide.
    crosslink_position : int
        Position of the crosslinker in the peptide sequence (1-based).

    Returns
    -------
    str
        The crosslinked peptide sequence in MS Annika format.

    Raises
    ------
    ValueError
        If the crosslink position is outside the peptide's length.

    Examples
    --------
    >>> from pyXLMS.exporter import get_msannika_crosslink_sequence
    >>> get_msannika_crosslink_sequence("PEPKTIDE", 4)
    'PEP[K]TIDE'

    >>> from pyXLMS.exporter import get_msannika_crosslink_sequence
    >>> get_msannika_crosslink_sequence("KPEPTIDE", 1)
    '[K]PEPTIDE'

    >>> from pyXLMS.exporter import get_msannika_crosslink_sequence
    >>> get_msannika_crosslink_sequence("PEPTIDEK", 8)
    'PEPTIDE[K]'
    """
    if crosslink_position < 1 or crosslink_position > len(peptide):
        raise ValueError(
            f"Crosslink position outside of range! Must be in range [1, {len(peptide)}]."
        )
    return f"{peptide[: crosslink_position - 1]}[{peptide[crosslink_position - 1]}]{peptide[crosslink_position:]}"


def __get_csm_td(value: Optional[bool]) -> str | None:
    r"""Helper function to get the [Alpha|Beta] T/D value.

    Parameters
    ----------
    value : bool, or None
        Decoy value of the crosslink-spectrum-match, should be either "alpha_decoy" or "beta_decoy" attribute.

    Returns
    -------
    str, or None
        If None was provided, None is returned. If a boolean is provided, returns "D" if True or "T" if False.

    Notes
    -----
    This function should not be called directly, it is called from ``__csms_to_msannika()``.
    """
    _ok = check_input(value, "value", bool) if value is not None else True
    if value is None:
        return None
    if value:
        return "D"
    return "T"


def __get_xl_isdecoy(
    alpha_decoy: Optional[bool], beta_decoy: Optional[bool]
) -> bool | None:
    r"""Helper function to get the Decoy value.

    Parameters
    ----------
    alpha_decoy : bool, or None
        Decoy value for the alpha peptide of the crosslink, should be "alpha_decoy" attribute.
    beta_decoy : bool, or None
        Decoy value for the beta peptide of the crosslink, should be "beta_decoy" attribute.

    Returns
    -------
    bool, or None
        If None was provided for any of the inputs, None is returned. Otherwise returns True if any of the inputs
        is True (= a decoy).

    Notes
    -----
    This function should not be called directly, it is called from ``__xls_to_msannika()``.
    """
    _ok = (
        check_input(alpha_decoy, "alpha_decoy", bool)
        if alpha_decoy is not None
        else None
    )
    _ok = (
        check_input(beta_decoy, "beta_decoy", bool) if beta_decoy is not None else None
    )
    if alpha_decoy is None or beta_decoy is None:
        return None
    return alpha_decoy or beta_decoy


def __csms_to_msannika(
    csms: List[Dict[str, Any]],
    filename: Optional[str],
    format: Literal["csv", "tsv", "xlsx"],
) -> pd.DataFrame:
    r"""Exports crosslink-spectrum-matches to MS Annika format.

    Parameters
    ----------
    csms : list of dict of str, any
        A list of crosslink-spectrum-matches.
    filename : str, or None
        If not None, the data will be written to a file with the specified filename.
    format : str, one of "csv", "tsv", or "xlsx"
        Format of the output file if filename is not None.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame in MS Annika format.

    Notes
    -----
    This function should not be called directly, it is called from ``to_msannika()``.
    """
    sequence = list()
    crosslink_type = list()
    sequence_a = list()
    crosslinker_position_a = list()
    accession_a = list()
    a_in_protein = list()
    score_alpha = list()
    alpha_td = list()
    sequence_b = list()
    crosslinker_position_b = list()
    accession_b = list()
    b_in_protein = list()
    score_beta = list()
    beta_td = list()
    combined_score = list()
    spectrum_file = list()
    first_scan = list()
    charge = list()
    rt_min = list()
    compensation_voltage = list()
    for csm in csms:
        sequence.append(f"{csm['alpha_peptide']}-{csm['beta_peptide']}")
        crosslink_type.append("Intra" if csm["crosslink_type"] == "intra" else "Inter")
        sequence_a.append(csm["alpha_peptide"])
        crosslinker_position_a.append(csm["alpha_peptide_crosslink_position"])
        accession_a.append(
            ";".join(csm["alpha_proteins"])
            if csm["alpha_proteins"] is not None
            else None
        )
        a_in_protein.append(
            ";".join([str(pos - 1) for pos in csm["alpha_proteins_peptide_positions"]])
            if csm["alpha_proteins_peptide_positions"] is not None
            else None
        )
        score_alpha.append(csm["alpha_score"])
        alpha_td.append(__get_csm_td(csm["alpha_decoy"]))
        sequence_b.append(csm["beta_peptide"])
        crosslinker_position_b.append(csm["beta_peptide_crosslink_position"])
        accession_b.append(
            ";".join(csm["beta_proteins"]) if csm["beta_proteins"] is not None else None
        )
        b_in_protein.append(
            ";".join([str(pos - 1) for pos in csm["beta_proteins_peptide_positions"]])
            if csm["beta_proteins_peptide_positions"] is not None
            else None
        )
        score_beta.append(csm["beta_score"])
        beta_td.append(__get_csm_td(csm["beta_decoy"]))
        combined_score.append(csm["score"])
        spectrum_file.append(csm["spectrum_file"])
        first_scan.append(csm["scan_nr"])
        charge.append(csm["charge"])
        rt_min.append(
            csm["retention_time"] / 60.0 if csm["retention_time"] is not None else None
        )
        compensation_voltage.append(csm["ion_mobility"])
    msannika_df = pd.DataFrame(
        {
            "Sequence": sequence,
            "Crosslink Type": crosslink_type,
            "Sequence A": sequence_a,
            "Crosslinker Position A": crosslinker_position_a,
            "Accession A": accession_a,
            "A in protein": a_in_protein,
            "Score Alpha": score_alpha,
            "Alpha T/D": alpha_td,
            "Sequence B": sequence_b,
            "Crosslinker Position B": crosslinker_position_b,
            "Accession B": accession_b,
            "B in protein": b_in_protein,
            "Score Beta": score_beta,
            "Beta T/D": beta_td,
            "Combined Score": combined_score,
            "Spectrum File": spectrum_file,
            "First Scan": first_scan,
            "Charge": charge,
            "RT [min]": rt_min,
            "Compensation Voltage": compensation_voltage,
        }
    )
    if filename is not None:
        if format == "csv":
            msannika_df.to_csv(__get_filename(filename, format), index=False)
        elif format == "tsv":
            msannika_df.to_csv(__get_filename(filename, format), sep="\t", index=False)
        else:
            msannika_df.to_excel(
                __get_filename(filename, format), engine="openpyxl", index=False
            )
    return msannika_df


def __xls_to_msannika(
    xls: List[Dict[str, Any]],
    filename: Optional[str],
    format: Literal["csv", "tsv", "xlsx"],
) -> pd.DataFrame:
    r"""Exports crosslinks to MS Annika format.

    Parameters
    ----------
    xls : list of dict of str, any
        A list of crosslinks.
    filename : str, or None
        If not None, the data will be written to a file with the specified filename.
    format : str, one of "csv", "tsv", or "xlsx"
        Format of the output file if filename is not None.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame in MS Annika format.

    Notes
    -----
    This function should not be called directly, it is called from ``to_msannika()``.
    """
    crosslink_type = list()
    sequence_a = list()
    position_a = list()
    accession_a = list()
    in_protein_a = list()
    sequence_b = list()
    position_b = list()
    accession_b = list()
    in_protein_b = list()
    best_csm_score = list()
    decoy = list()
    for xl in xls:
        crosslink_type.append("Intra" if xl["crosslink_type"] == "intra" else "Inter")
        sequence_a.append(
            get_msannika_crosslink_sequence(
                xl["alpha_peptide"], xl["alpha_peptide_crosslink_position"]
            )
        )
        position_a.append(xl["alpha_peptide_crosslink_position"])
        accession_a.append(
            ";".join(xl["alpha_proteins"]) if xl["alpha_proteins"] is not None else None
        )
        in_protein_a.append(
            ";".join([str(pos) for pos in xl["alpha_proteins_crosslink_positions"]])
            if xl["alpha_proteins_crosslink_positions"] is not None
            else None
        )
        sequence_b.append(
            get_msannika_crosslink_sequence(
                xl["beta_peptide"], xl["beta_peptide_crosslink_position"]
            )
        )
        position_b.append(xl["beta_peptide_crosslink_position"])
        accession_b.append(
            ";".join(xl["beta_proteins"]) if xl["beta_proteins"] is not None else None
        )
        in_protein_b.append(
            ";".join([str(pos) for pos in xl["beta_proteins_crosslink_positions"]])
            if xl["beta_proteins_crosslink_positions"] is not None
            else None
        )
        best_csm_score.append(xl["score"])
        decoy.append(__get_xl_isdecoy(xl["alpha_decoy"], xl["beta_decoy"]))
    msannika_df = pd.DataFrame(
        {
            "Crosslink Type": crosslink_type,
            "Sequence A": sequence_a,
            "Position A": position_a,
            "Accession A": accession_a,
            "In protein A": in_protein_a,
            "Sequence B": sequence_b,
            "Position B": position_b,
            "Accession B": accession_b,
            "In protein B": in_protein_b,
            "Best CSM Score": best_csm_score,
            "Decoy": decoy,
        }
    )
    if filename is not None:
        if format == "csv":
            msannika_df.to_csv(__get_filename(filename, format), index=False)
        elif format == "tsv":
            msannika_df.to_csv(__get_filename(filename, format), sep="\t", index=False)
        else:
            msannika_df.to_excel(
                __get_filename(filename, format), engine="openpyxl", index=False
            )
    return msannika_df


def to_msannika(
    data: List[Dict[str, Any]],
    filename: Optional[str] = None,
    format: Literal["csv", "tsv", "xlsx"] = "csv",
) -> pd.DataFrame:
    r"""Exports a list of crosslinks or crosslink-spectrum-matches to MS Annika format.

    Exports a list of crosslinks or crosslink-spectrum-matches to MS Annika format. This might be useful
    for tools that support MS Annika input but are not supported by pyXLMS (yet).

    Parameters
    ----------
    data : list of dict of str, any
        A list of crosslinks or crosslink-spectrum-matches.
    filename : str, or None, default = None
        If not None, the exported data will be written to a file with the specified filename.
    format : str, one of "csv", "tsv", or "xlsx", default = "csv"
        File format of the exported file if filename is not None.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame containing crosslinks or crosslink-spectrum-matches in MS Annika format.

    Raises
    ------
    TypeError
        If a wrong data type is provided.
    TypeError
        If data contains elements of mixed data type.
    TypeError
        If parameter format is not one of 'csv', 'tsv' or 'xlsx'.
    ValueError
        If the provided data contains no elements.

    Warnings
    --------
    The MS Annika exporter will not check if all necessary information is available for the exported
    crosslinks or crosslink-spectrum-matches. If a value is not available it will be denoted as a missing
    value in the dataframe and exported file. Please make sure all necessary information is available
    before using the exported file with another tool! Please also note that modifications are not exported,
    for modification down-stream analysis please refer to ``transform.to_proforma()`` or
    ``transform.to_dataframe()``!

    Examples
    --------
    >>> from pyXLMS.exporter import to_msannika
    >>> from pyXLMS.data import create_crosslink_min
    >>> xl1 = create_crosslink_min("KPEPTIDE", 1, "PKEPTIDE", 2)
    >>> xl2 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    >>> crosslinks = [xl1, xl2]
    >>> to_msannika(crosslinks)
      Crosslink Type  Sequence A  Position A Accession A In protein A  Sequence B  Position B Accession B In protein B Best CSM Score Decoy
    0          Inter  [K]PEPTIDE           1        None         None  P[K]EPTIDE           2        None         None           None  None
    1          Inter  PE[K]PTIDE           3        None         None  PEP[K]TIDE           4        None         None           None  None

    >>> from pyXLMS.exporter import to_msannika
    >>> from pyXLMS.data import create_crosslink_min
    >>> xl1 = create_crosslink_min("KPEPTIDE", 1, "PKEPTIDE", 2)
    >>> xl2 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    >>> crosslinks = [xl1, xl2]
    >>> df = to_msannika(crosslinks, filename="crosslinks.csv", format="csv")

    >>> from pyXLMS.exporter import to_msannika
    >>> from pyXLMS.data import create_csm_min
    >>> csm1 = create_csm_min("KPEPTIDE", 1, "PKEPTIDE", 2, "RUN_1", 1)
    >>> csm2 = create_csm_min("PEKPTIDE", 3, "PEPKTIDE", 4, "RUN_1", 2)
    >>> csms = [csm1, csm2]
    >>> to_msannika(csms)
                Sequence Crosslink Type Sequence A  Crosslinker Position A  ... First Scan Charge RT [min] Compensation Voltage
    0  KPEPTIDE-PKEPTIDE          Inter   KPEPTIDE                       1  ...          1   None     None                 None
    1  PEKPTIDE-PEPKTIDE          Inter   PEKPTIDE                       3  ...          2   None     None                 None
    [2 rows x 20 columns]

    >>> from pyXLMS.exporter import to_msannika
    >>> from pyXLMS.data import create_csm_min
    >>> csm1 = create_csm_min("KPEPTIDE", 1, "PKEPTIDE", 2, "RUN_1", 1)
    >>> csm2 = create_csm_min("PEKPTIDE", 3, "PEPKTIDE", 4, "RUN_1", 2)
    >>> csms = [csm1, csm2]
    >>> df = to_msannika(csms, filename="csms.csv", format="csv")
    """
    _ok = check_input(data, "data", list, dict)
    _ok = check_input(filename, "filename", str) if filename is not None else True
    _ok = check_input(format, "format", str)
    if format not in ["csv", "tsv", "xlsx"]:
        raise TypeError("Parameter 'format' has to be one of 'csv', 'tsv', or 'xlsx'!")
    if len(data) == 0:
        raise ValueError(
            "Provided data does not contain any crosslinks or crosslink-spectrum-matches!"
        )
    if "data_type" not in data[0] or data[0]["data_type"] not in [
        "crosslink",
        "crosslink-spectrum-match",
    ]:
        raise TypeError(
            "Unsupported data type for input data! Parameter data has to be a list of crosslink or crosslink-spectrum-match!"
        )
    if not assert_data_type_same(data):
        raise TypeError("Not all elements in data have the same data type!")
    if data[0]["data_type"] == "crosslink":
        return __xls_to_msannika(data, filename, format)
    return __csms_to_msannika(data, filename, format)



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import pandas as pd

from ..data import check_input
from ..data import create_crosslink_from_csm
from ..transform.util import get_available_keys
from ..transform.filter import filter_target_decoy
from .to_msannika import to_msannika

from typing import Optional
from typing import Dict
from typing import Any
from typing import List


def to_impxfdr(
    data: List[Dict[str, Any]],
    filename: Optional[str],
    targets_only: bool = True,
) -> pd.DataFrame:
    r"""Exports a list of crosslinks or crosslink-spectrum-matches to IMP-X-FDR format.

    Exports a list of crosslinks or crosslink-spectrum-matches to IMP-X-FDR format for benchmarking purposes.
    The tool IMP-X-FDR is available from
    `github.com/vbc-proteomics-org/imp-x-fdr <https://github.com/vbc-proteomics-org/imp-x-fdr>`_.
    We recommend using version 1.1.0 and selecting "MS Annika" as input file format for the here exported file.
    A slightly modified version is available from
    `github.com/hgb-bin-proteomics/MSAnnika_NC_Results <https://github.com/hgb-bin-proteomics/MSAnnika_NC_Results/blob/master/Peplib_Beveridge/MS_Annika/Tools/IMP-X-FDR.v1.1.0.zip>`_.
    This version contains a few bug fixes and was used for the MS Annika 2.0 and MS Annika 3.0 publications.
    Requires that ``alpha_proteins``, ``beta_proteins``, ``alpha_proteins_crosslink_positions`` and ``beta_proteins_crosslink_positions`` fields
    are set for crosslinks and crosslink-spectrum-matches.

    Parameters
    ----------
    data : list of dict of str, any
        A list of crosslinks or crosslink-spectrum-matches.
    filename : str, or None, default = None
        If not None, the exported data will be written to a file with the specified filename.
        The filename should end in ".xlsx" as the file is exported to Microsoft Excel file format.
    targets_only : bool, default = True
        Whether or not only target crosslinks or crosslink-spectrum-matches should be exported. For
        benchmarking purposes this is usually the case. If the crosslinks or crosslink-spectrum-matches
        do not contain target-decoy labels this should be set to False.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame containing crosslinks or crosslink-spectrum-matches in IMP-X-FDR format.

    Raises
    ------
    TypeError
        If a wrong data type is provided.
    TypeError
        If data contains elements of mixed data type.
    ValueError
        If the provided data contains no elements or if none of the data has target-decoy labels
        and parameter 'targets_only' is set to True.
    RuntimeError
        If not all of the required information is present in the input data.

    Examples
    --------
    >>> from pyXLMS.exporter import to_impxfdr
    >>> from pyXLMS.parser import read
    >>> pr = read(
    ...     "data/xi/1perc_xl_boost_Links_xiFDR2.2.1.csv",
    ...     engine="xiSearch/xiFDR",
    ...     crosslinker="DSS",
    ... )
    >>> crosslinks = pr["crosslinks"]
    >>> to_impxfdr(crosslinks, filename="crosslinks.xlsx")
        Crosslink Type             Sequence A  Position A Accession A In protein A  ... Position B  Accession B In protein B Best CSM Score  Decoy
    0            Intra          VVDELV[K]VMGR           7        Cas9          753  ...          7         Cas9          753         40.679  False
    1            Intra  MLASAGELQ[K]GNELALPSK          10        Cas9          753  ...          7         Cas9         1226         40.231  False
    2            Intra        MDGTEELLV[K]LNR          10        Cas9          396  ...         10         Cas9          396         39.582  False
    3            Intra         MTNFD[K]NLPNEK           6        Cas9          965  ...          2         Cas9          504         35.880  False
    4            Intra             DFQFY[K]VR           6        Cas9          978  ...          4         Cas9         1028         35.281  False
    ..             ...                    ...         ...         ...          ...  ...        ...          ...          ...            ...    ...
    220          Intra        LP[K]YSLFELENGR           3        Cas9          866  ...          3         Cas9         1204          9.877  False
    221          Intra               D[K]QSGK           2        Cas9          677  ...          2         Cas9          677          9.702  False
    222          Intra               AGFI[K]R           5        Cas9          922  ...         11         Cas9          881          9.666  False
    223          Intra                E[K]IEK           2        Cas9          443  ...          1         Cas9          562          9.656  False
    224          Intra                LS[K]SR           3        Cas9          222  ...          3         Cas9          222          9.619  False
    [225 rows x 11 columns]

    >>> from pyXLMS.exporter import to_impxfdr
    >>> from pyXLMS.parser import read
    >>> pr = read(
    ...     "data/xi/1perc_xl_boost_CSM_xiFDR2.2.1.csv",
    ...     engine="xiSearch/xiFDR",
    ...     crosslinker="DSS",
    ... )
    >>> csms = pr["crosslink-spectrum-matches"]
    >>> to_impxfdr(csms, filename="csms.xlsx")
        Crosslink Type          Sequence A  Position A Accession A In protein A  ... Position B  Accession B In protein B Best CSM Score  Decoy
    0            Intra  [K]IECFDSVEISGVEDR           1        Cas9          575  ...          1         Cas9          575         27.268  False
    1            Intra       LVDSTD[K]ADLR           7        Cas9          152  ...         11         Cas9          881         26.437  False
    2            Intra     GGLSELD[K]AGFIK           8        Cas9          917  ...          8         Cas9          917         26.134  False
    3            Intra       LVDSTD[K]ADLR           7        Cas9          152  ...          7         Cas9          152         25.804  False
    4            Intra       VVDELV[K]VMGR           7        Cas9          753  ...          7         Cas9          753         24.861  False
    ..             ...                 ...         ...         ...          ...  ...        ...          ...          ...            ...    ...
    406          Intra          [K]GILQTVK           1        Cas9          739  ...          3         Cas9          222          6.977  False
    407          Intra          QQLPE[K]YK           6        Cas9          350  ...          6         Cas9          350          6.919  False
    408          Intra           ESILP[K]R           6        Cas9         1117  ...          7         Cas9         1035          6.853  False
    409          Intra             LS[K]SR           3        Cas9          222  ...          2         Cas9          884          6.809  False
    410          Intra     QIT[K]HVAQILDSR           4        Cas9          933  ...          6         Cas9          350          6.808  False
    [411 rows x 11 columns]
    """
    _ok = check_input(data, "data", list, dict)
    _ok = check_input(filename, "filename", str) if filename is not None else True
    _ok = check_input(targets_only, "targets_only", bool)
    if targets_only:
        data = filter_target_decoy(data)["Target-Target"]
    if len(data) == 0:
        if targets_only:
            raise ValueError(
                "Provided data does not contain any crosslinks or crosslink-spectrum-matches after filtering for targets only!"
            )
        else:
            raise ValueError(
                "Provided data does not contain any crosslinks or crosslink-spectrum-matches!"
            )
    if "data_type" not in data[0] or data[0]["data_type"] not in [
        "crosslink",
        "crosslink-spectrum-match",
    ]:
        raise TypeError(
            "Unsupported data type for input data! Parameter data has to be a list of crosslink or crosslink-spectrum-match!"
        )
    available_keys = get_available_keys(data)
    if (
        not available_keys["alpha_proteins"]
        or not available_keys["alpha_proteins_crosslink_positions"]
        or not available_keys["beta_proteins"]
        or not available_keys["beta_proteins_crosslink_positions"]
    ):
        raise RuntimeError(
            "Can't export to IMP-X-FDR because not all necessary information is available!"
        )
    if data[0]["data_type"] == "crosslink":
        return to_msannika(data, filename, format="xlsx")
    return to_msannika(
        [create_crosslink_from_csm(csm) for csm in data], filename, format="xlsx"
    )



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import io
import warnings
import pandas as pd
from tqdm import tqdm

from ..data import check_input
from ..data import create_csm
from ..data import create_crosslink
from ..data import create_parser_result
from ..constants import MODIFICATIONS
from .util import format_sequence
from .util import __serialize_pandas_series

from typing import Optional
from typing import BinaryIO
from typing import Dict
from typing import Any
from typing import Tuple
from typing import List
from typing import Callable

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def __parse_modifications_from_plink_modifications_str(
    seq: str,
    mod_str: Optional[str | float],
    crosslinker: str,
    modifications: Dict[str, float] = MODIFICATIONS,
    verbose: Literal[0, 1, 2] = 1,
) -> Tuple[Dict[int, Tuple[str, float]], Dict[int, Tuple[str, float]]]:
    r"""Parse post-translational-modifications from a pLink modification string.

    Parses post-translational-modifications (PTMs) from a pLink modification string,
    for example "Carbamidomethyl[C](4);Oxidation[M](23)".

    Parameters
    ----------
    seq : str
        The pLink crosslink sequence string.
    mod_str : str, float, or None
        The pLink modification value, as string or float. Can be None.
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    modifications: dict of str, float, default = ``constants.MODIFICATIONS``
        Mapping of modification names to modification masses.
    verbose : 0, 1, or 2, default = 1
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    tuple of dict of int, tuple
        The ``pyXLMS`` specific modification objects, dictionaries that map positions to their corresponding modifications and their
        monoisotopic masses. The first object (index 0) corresponds to the modifications of the first peptide, the second object (index 1)
        corresponds to the modifications of the second peptide.

    Raises
    ------
    RuntimeError
        If multiple modifications on the same residue are parsed (only for ``verbose = 2``).
    KeyError
        If an unknown modification is encountered.

    Notes
    -----
    This function should not be called directly, it is called from ``read_plink()``.
    """
    modifications_a = dict()
    modifications_b = dict()
    xl_pos_a = int(seq.split("-")[0].split("(")[1].split(")")[0])
    xl_pos_b = int(seq.split("-")[1].split("(")[1].split(")")[0])
    if crosslinker in modifications:
        modifications_a[xl_pos_a] = (crosslinker, modifications[crosslinker])
        modifications_b[xl_pos_b] = (crosslinker, modifications[crosslinker])
    else:
        raise KeyError(
            f"Key {crosslinker} not found in parameter 'modifications'. Are you missing a modification?"
        )
    if mod_str is None:
        return (modifications_a, modifications_b)
    if isinstance(mod_str, float) and pd.isna(mod_str):
        return (modifications_a, modifications_b)
    mod_str = str(mod_str).strip()
    if mod_str == "nan":
        return (modifications_a, modifications_b)
    mods = mod_str.split(";")
    for mod in mods:
        mod_desc = mod.split("[")[0].strip()
        if mod_desc not in modifications:
            raise KeyError(
                f"Key {mod_desc} not found in parameter 'modifications'. Are you missing a modification?"
            )
        mod_pos = int(mod.split("(")[1].split(")")[0])
        if mod_pos > len(seq.split("-")[0]):
            mod_pos = mod_pos - len(seq.split("-")[0])
            if mod_pos in modifications_b:
                if verbose == 2:
                    raise RuntimeError(
                        f"Modification at position {mod_pos} already exists!"
                    )
                if verbose == 1:
                    warnings.warn(
                        RuntimeWarning(
                            f"Modification at position {mod_pos} already exists!"
                        )
                    )
                t1 = modifications_b[mod_pos][0] + "," + mod_desc
                t2 = modifications_b[mod_pos][1] + modifications[mod_desc]
                modifications_b[mod_pos] = (t1, t2)
            else:
                modifications_b[mod_pos] = (mod_desc, modifications[mod_desc])
        else:
            if mod_pos in modifications_a:
                if verbose == 2:
                    raise RuntimeError(
                        f"Modification at position {mod_pos} already exists!"
                    )
                if verbose == 1:
                    warnings.warn(
                        RuntimeWarning(
                            f"Modification at position {mod_pos} already exists!"
                        )
                    )
                t1 = modifications_a[mod_pos][0] + "," + mod_desc
                t2 = modifications_a[mod_pos][1] + modifications[mod_desc]
                modifications_a[mod_pos] = (t1, t2)
            else:
                modifications_a[mod_pos] = (mod_desc, modifications[mod_desc])
    return (modifications_a, modifications_b)


def __parse_proteins_and_position_from_plink(
    seq: str,
    proteins: str,
) -> Dict[str, Any]:
    r"""Parses proteins and positions from pLink results.

    Parses proteins, as well as peptide and crosslink positions from a pLink crosslink sequence
    and protein string.

    Parameters
    ----------
    seq : str
        The pLink crosslink sequence string.
    proteins : str
        The pLink proteins string.

    Returns
    -------
    dict of str, Any
        A dictionary with the following keys and information:
        ``xl_pos_a``, ``proteins_a``, ``proteins_a_xl_positions``, ``proteins_a_pep_positions``,
        ``xl_pos_b``, ``proteins_b``, ``proteins_b_xl_positions``, ``proteins_b_pep_positions``.

    Notes
    -----
    This function should not be called directly, it is called from ``read_plink()``.
    """
    xl_pos_a = int(seq.split("-")[0].split("(")[1].split(")")[0])
    xl_pos_b = int(seq.split("-")[1].split("(")[1].split(")")[0])
    # proteins a
    proteins_set_a = set()
    proteins_a = list()
    proteins_a_xl_positions = list()
    proteins_a_pep_positions = list()
    # proteins b
    proteins_set_b = set()
    proteins_b = list()
    proteins_b_xl_positions = list()
    proteins_b_pep_positions = list()
    # find unique
    proteins = proteins.strip().rstrip("/")
    for protein_pair in proteins.split("/"):
        protein_a = protein_pair.split("-")[0].strip()
        protein_b = protein_pair.split("-")[1].strip()
        proteins_set_a.add(protein_a)
        proteins_set_b.add(protein_b)
    # get proteins a
    for protein in sorted(proteins_set_a):
        acc = protein.split("(")[0]
        pos = int(protein.split("(")[1].split(")")[0])
        proteins_a.append(acc)
        proteins_a_xl_positions.append(pos)
        proteins_a_pep_positions.append(pos - xl_pos_a + 1)
    # get proteins b
    for protein in sorted(proteins_set_b):
        acc = protein.split("(")[0]
        pos = int(protein.split("(")[1].split(")")[0])
        proteins_b.append(acc)
        proteins_b_xl_positions.append(pos)
        proteins_b_pep_positions.append(pos - xl_pos_b + 1)
    return {
        "xl_pos_a": xl_pos_a,
        "proteins_a": proteins_a,
        "proteins_a_xl_positions": proteins_a_xl_positions,
        "proteins_a_pep_positions": proteins_a_pep_positions,
        "xl_pos_b": xl_pos_b,
        "proteins_b": proteins_b,
        "proteins_b_xl_positions": proteins_b_xl_positions,
        "proteins_b_pep_positions": proteins_b_pep_positions,
    }


def __read_plink_cross_linked_peptides_file(
    file: str | BinaryIO, sep: str = ",", decimal: str = "."
) -> pd.DataFrame:
    r"""Reads a pLink cross-linked peptides file into a pandas DataFrame.

    Reads a pLink cross-linked peptides file into a pandas DataFrame. Reading
    skips all lines starting with ``sep``.

    Parameters
    ----------
    file : str, or BinaryIO
        The name/path of the pLink result file or a file-like object/stream.
    sep : str, default = ","
        Seperator used in the ``.csv`` file.
    decimal : str, default = "."
        Character to recognize as decimal point.

    Returns
    -------
    pd.DataFrame
        The parsed cross-linked peptides.

    Raises
    ------
    RuntimeError
        If the file could not be parsed.
    ValueError
        If the file is not a pLink cross-linked peptides file.

    Notes
    -----
    This function should not be called directly, it is called from ``read_plink()`` and ``detect_plink_filetype()``.
    """
    parsed_lines = None
    if isinstance(file, str):
        with open(file, "r", encoding="utf-8") as f:
            parsed_lines = f.readlines()
            f.close()
    else:
        parsed_lines = file.readlines()
        file.seek(0)
    if parsed_lines is None:
        raise RuntimeError("Something went wrong while parsing the file!")
    lines = list()
    for line in parsed_lines:
        preprocessed_line = str(line).strip()
        if not preprocessed_line.startswith(sep):
            lines.append(preprocessed_line)
    df = pd.read_csv(
        io.StringIO("\n".join(lines)), sep=sep, decimal=decimal, low_memory=False
    )
    colnames = df.columns.values.tolist()
    if not (
        "Peptide_Order" in colnames
        and "Peptide" in colnames
        and "Peptide_Mass" in colnames
        and "Modifications" in colnames
        and "Proteins" in colnames
        and "Protein_Type" in colnames
    ):
        raise ValueError(
            "The provided file seems not to be a pLink cross-linked peptides file!"
        )
    return df


def parse_spectrum_file_from_plink(title: str) -> str:
    r"""Parse the spectrum file name from a spectrum title.

    Parameters
    ----------
    title : str
        The spectrum title.

    Returns
    -------
    str
        The spectrum file name.

    Examples
    --------
    >>> from pyXLMS.parser import parse_spectrum_file_from_plink
    >>> parse_spectrum_file_from_plink(
    ...     "XLpeplib_Beveridge_QEx-HFX_DSS_R1.20588.20588.3.0.dta"
    ... )
    'XLpeplib_Beveridge_QEx-HFX_DSS_R1'
    """
    return str(title).split(".")[0].strip()


def parse_scan_nr_from_plink(title: str) -> int:
    r"""Parse the scan number from a spectrum title.

    Parameters
    ----------
    title : str
        The spectrum title.

    Returns
    -------
    int
        The scan number.

    Examples
    --------
    >>> from pyXLMS.parser import parse_scan_nr_from_plink
    >>> parse_scan_nr_from_plink(
    ...     "XLpeplib_Beveridge_QEx-HFX_DSS_R1.20588.20588.3.0.dta"
    ... )
    20588
    """
    return int(str(title).split(".")[1])


def detect_plink_filetype(
    file: str | BinaryIO, sep: str = ",", decimal: str = "."
) -> Literal["crosslinks", "crosslink-spectrum-matches"]:
    r"""Detects the pLink-related file type of the data.

    Detects whether the input data is a pLink "\*cross-linked_peptides.csv" file or
    a pLink "\*cross-linked_spectra.csv" file.

    Parameters
    ----------
    file : str, or BinaryIO
        The name/path of the pLink result file or a file-like object/stream.
    sep : str, default = ","
        Seperator used in the ``.csv`` file.
    decimal : str, default = "."
        Character to recognize as decimal point.

    Returns
    -------
    str
        Returns "crosslinks" if ``file`` is a "\*cross-linked_peptides.csv" or
        "crosslink-spectrum-matches" if ``file`` is a "\*cross-linked_spectra.csv".

    Raises
    ------
    RuntimeError
        If the file could not be parsed.
    RuntimeError
        If the file does not contain any data.
    ValueError
        If the file does not match any of the supported pLink input files.

    Examples
    --------
    >>> from pyXLMS.parser import detect_plink_filetype
    >>> detect_plink_filetype(
    ...     "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_peptides.csv"
    ... )
    'crosslinks'

    >>> from pyXLMS.parser import detect_plink_filetype
    >>> detect_plink_filetype(
    ...     "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_spectra.csv"
    ... )
    'crosslink-spectrum-matches'
    """
    # parse file
    parsed_lines = None
    if isinstance(file, str):
        with open(file, "r", encoding="utf-8") as f:
            parsed_lines = f.readlines()
            f.close()
    else:
        parsed_lines = file.readlines()
        file.seek(0)
    if parsed_lines is None:
        raise RuntimeError("Something went wrong while parsing the file!")
    lines = list()
    # filter empty lines
    for line in parsed_lines:
        if len(str(line).strip()) > 0:
            lines.append(str(line).strip())
    # checks
    if len(lines) == 0:
        raise RuntimeError("The provided file does not contain any data!")
    if len(lines) == 1:
        df = pd.read_csv(file, sep=sep, decimal=decimal, low_memory=False)
        if not isinstance(file, str):
            file.seek(0)
        colnames = df.columns.values.tolist()
        if (
            "Peptide" in colnames
            and "Modifications" in colnames
            and "Linker" in colnames
            and "Proteins" in colnames
            and "Score" in colnames
            and "Title" in colnames
            and "Charge" in colnames
            and "Evalue" in colnames
            and "Alpha_Evalue" in colnames
            and "Beta_Evalue" in colnames
        ):
            return "crosslink-spectrum-matches"
        else:
            raise ValueError(
                "The provided file seems not to be one of the support pLink input files!"
            )
    if lines[1].startswith(sep):
        _ = __read_plink_cross_linked_peptides_file(file, sep=sep, decimal=decimal)
        return "crosslinks"
    df = pd.read_csv(file, sep=sep, decimal=decimal, low_memory=False)
    if not isinstance(file, str):
        file.seek(0)
    colnames = df.columns.values.tolist()
    if not (
        "Peptide" in colnames
        and "Modifications" in colnames
        and "Linker" in colnames
        and "Proteins" in colnames
        and "Score" in colnames
        and "Title" in colnames
        and "Charge" in colnames
        and "Evalue" in colnames
        and "Alpha_Evalue" in colnames
        and "Beta_Evalue" in colnames
    ):
        raise ValueError(
            "The provided file seems not to be one of the support pLink input files!"
        )
    return "crosslink-spectrum-matches"


def read_plink(
    files: str | List[str] | BinaryIO,
    spectrum_file_parser: Optional[Callable[[str], str]] = None,
    scan_nr_parser: Optional[Callable[[str], int]] = None,
    decoy_prefix: str = "REV_",
    parse_modifications: bool = True,
    modifications: Dict[str, float] = MODIFICATIONS,
    sep: str = ",",
    decimal: str = ".",
    verbose: Literal[0, 1, 2] = 1,
) -> Dict[str, Any]:
    r"""Read a pLink result file.

    Reads a pLink crosslink-spectrum-matches result file "\*cross-linked_spectra.csv"
    in ``.csv`` (comma delimited) format or a pLink crosslinks result file
    "\*cross-linked_peptides.csv" in ``.csv`` (comma delimited) format and
    returns a ``parser_result``.

    Parameters
    ----------
    files : str, list of str, or file stream
        The name/path of the pLink result file(s) or a file-like object/stream.
    spectrum_file_parser: callable, or None, default = None
        A function that parses the spectrum file name from spectrum titles. If None (default)
        the function ``parse_spectrum_file_from_plink()`` is used.
    scan_nr_parser : callable, or None, default = None
        A function that parses the scan number from spectrum titles. If None (default)
        the function ``parse_scan_nr_from_plink()`` is used.
    decoy_prefix : str, default = "REV\_"
        The prefix that indicates that a protein is from the decoy database.
    parse_modifications : bool, default = True
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modifications' parameter.
    modifications: dict of str, float, default = ``constants.MODIFICATIONS``
        Mapping of modification names to modification masses.
    sep : str, default = ","
        Seperator used in the ``.csv`` file.
    decimal : str, default = "."
        Character to recognize as decimal point.
    verbose : 0, 1, or 2, default = 1
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    dict
        The ``parser_result`` object containing all parsed information.

    Raises
    ------
    RuntimeError
        If the file(s) could not be read or if the file(s) contain no crosslink-spectrum-matches.
    TypeError
        If parameter verbose was not set correctly.

    Notes
    -----
    Uses ``Score`` as the score of the crosslink-spectrum-match, all other scores are ``None``.

    Warnings
    --------
    Target and decoy information is derived based off the protein accession and parameter ``decoy_prefix``.
    By default, pLink only reports target matches that are above the desired FDR.

    Examples
    --------
    >>> from pyXLMS.parser import read_plink
    >>> csms = read_plink(
    ...     "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_spectra.csv"
    ... )

    >>> from pyXLMS.parser import read_plink
    >>> crosslinks = read_plink(
    ...     "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_peptides.csv"
    ... )
    """
    ## check input
    _ok = (
        check_input(spectrum_file_parser, "spectrum_file_parser", Callable)
        if spectrum_file_parser is not None
        else True
    )
    _ok = (
        check_input(scan_nr_parser, "scan_nr_parser", Callable)
        if scan_nr_parser is not None
        else True
    )
    _ok = check_input(decoy_prefix, "decoy_prefix", str)
    _ok = check_input(parse_modifications, "parse_modifications", bool)
    _ok = check_input(modifications, "modifications", dict, float)
    _ok = check_input(sep, "sep", str)
    _ok = check_input(decimal, "decimal", str)
    _ok = check_input(verbose, "verbose", int)
    if verbose not in [0, 1, 2]:
        raise TypeError("Verbose level has to be one of 0, 1, or 2!")

    ## set default parsers
    if spectrum_file_parser is None:
        spectrum_file_parser = parse_spectrum_file_from_plink
    if scan_nr_parser is None:
        scan_nr_parser = parse_scan_nr_from_plink

    ## data structures
    csms = list()
    crosslinks = list()

    ## handle input
    if not isinstance(files, list):
        inputs = [files]
    else:
        inputs = files

    ## process data
    for input in inputs:
        if detect_plink_filetype(input, sep=sep, decimal=decimal) == "crosslinks":
            data = __read_plink_cross_linked_peptides_file(
                input, sep=sep, decimal=decimal
            )
            for i, row in tqdm(
                data.iterrows(), total=data.shape[0], desc="Reading pLink crosslinks..."
            ):
                parsed_positions = __parse_proteins_and_position_from_plink(
                    seq=str(row["Peptide"]).strip(),
                    proteins=str(row["Proteins"]).strip(),
                )
                # create csm
                crosslink = create_crosslink(
                    peptide_a=format_sequence(
                        str(row["Peptide"]).split("-")[0].split("(")[0].strip()
                    ),
                    xl_position_peptide_a=parsed_positions["xl_pos_a"],
                    proteins_a=[
                        protein_a.strip()
                        if protein_a.strip()[: len(decoy_prefix)] != decoy_prefix
                        else protein_a.strip()[len(decoy_prefix) :]
                        for protein_a in parsed_positions["proteins_a"]
                    ],
                    xl_position_proteins_a=parsed_positions["proteins_a_xl_positions"],
                    decoy_a=decoy_prefix in " ".join(parsed_positions["proteins_a"]),
                    peptide_b=format_sequence(
                        str(row["Peptide"]).split("-")[1].split("(")[0].strip()
                    ),
                    xl_position_peptide_b=parsed_positions["xl_pos_b"],
                    proteins_b=[
                        protein_b.strip()
                        if protein_b.strip()[: len(decoy_prefix)] != decoy_prefix
                        else protein_b.strip()[len(decoy_prefix) :]
                        for protein_b in parsed_positions["proteins_b"]
                    ],
                    xl_position_proteins_b=parsed_positions["proteins_b_xl_positions"],
                    decoy_b=decoy_prefix in " ".join(parsed_positions["proteins_b"]),
                    score=None,
                    additional_information={"source": __serialize_pandas_series(row)},
                )
                crosslinks.append(crosslink)
        else:
            data = pd.read_csv(input, sep=sep, decimal=decimal, low_memory=False)
            for i, row in tqdm(
                data.iterrows(), total=data.shape[0], desc="Reading pLink CSMs..."
            ):
                # pre information
                parsed_modifications = (
                    __parse_modifications_from_plink_modifications_str(
                        seq=str(row["Peptide"]).strip(),
                        mod_str=row["Modifications"],  # pyright: ignore [reportArgumentType]
                        crosslinker=str(row["Linker"]).strip(),
                        modifications=modifications,
                        verbose=verbose,
                    )
                    if parse_modifications
                    else None
                )
                parsed_positions = __parse_proteins_and_position_from_plink(
                    seq=str(row["Peptide"]).strip(),
                    proteins=str(row["Proteins"]).strip(),
                )
                # create csm
                csm = create_csm(
                    peptide_a=format_sequence(
                        str(row["Peptide"]).split("-")[0].split("(")[0].strip()
                    ),
                    modifications_a=parsed_modifications[0]
                    if parsed_modifications is not None
                    else None,
                    xl_position_peptide_a=parsed_positions["xl_pos_a"],
                    proteins_a=[
                        protein_a.strip()
                        if protein_a.strip()[: len(decoy_prefix)] != decoy_prefix
                        else protein_a.strip()[len(decoy_prefix) :]
                        for protein_a in parsed_positions["proteins_a"]
                    ],
                    xl_position_proteins_a=parsed_positions["proteins_a_xl_positions"],
                    pep_position_proteins_a=parsed_positions[
                        "proteins_a_pep_positions"
                    ],
                    score_a=None,
                    decoy_a=decoy_prefix in " ".join(parsed_positions["proteins_a"]),
                    peptide_b=format_sequence(
                        str(row["Peptide"]).split("-")[1].split("(")[0].strip()
                    ),
                    modifications_b=parsed_modifications[1]
                    if parsed_modifications is not None
                    else None,
                    xl_position_peptide_b=parsed_positions["xl_pos_b"],
                    proteins_b=[
                        protein_b.strip()
                        if protein_b.strip()[: len(decoy_prefix)] != decoy_prefix
                        else protein_b.strip()[len(decoy_prefix) :]
                        for protein_b in parsed_positions["proteins_b"]
                    ],
                    xl_position_proteins_b=parsed_positions["proteins_b_xl_positions"],
                    pep_position_proteins_b=parsed_positions[
                        "proteins_b_pep_positions"
                    ],
                    score_b=None,
                    decoy_b=decoy_prefix in " ".join(parsed_positions["proteins_b"]),
                    score=float(row["Score"]),
                    spectrum_file=spectrum_file_parser(str(row["Title"]).strip()),
                    scan_nr=scan_nr_parser(str(row["Title"]).strip()),
                    charge=int(row["Charge"]),
                    rt=None,
                    im_cv=None,
                    additional_information={
                        "source": __serialize_pandas_series(row),
                        "Evalue": float(row["Evalue"]),
                        "Alpha_Evalue": float(row["Alpha_Evalue"]),
                        "Beta_Evalue": float(row["Beta_Evalue"]),
                    },
                )
                csms.append(csm)
    ## check results
    if len(crosslinks) + len(csms) == 0:
        raise RuntimeError(
            "No crosslink-spectrum-matches or crosslinks were parsed! If this is unexpected, please file a bug report!"
        )
    ## return parser result
    return create_parser_result(
        search_engine="pLink",
        csms=csms if len(csms) > 0 else None,
        crosslinks=crosslinks if len(crosslinks) > 0 else None,
    )



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

__all__ = [
    "to_xmas",
    "to_xlinkdb",
    "to_impxfdr",
    "to_msannika",
    "get_msannika_crosslink_sequence",
    "to_pyxlinkviewer",
    "to_xlmstools",
    "to_xinet",
    "to_xiview",
    "to_xifdr",
    "to_alphalink2",
    "to_proxl",
]

from .to_xmas import to_xmas
from .to_xlinkdb import to_xlinkdb
from .to_impxfdr import to_impxfdr
from .to_msannika import to_msannika
from .to_msannika import get_msannika_crosslink_sequence
from .to_pyxlinkviewer import to_pyxlinkviewer
from .to_xlmstools import to_xlmstools
from .to_xinet import to_xinet
from .to_xiview import to_xiview
from .to_xifdr import to_xifdr
from .to_alphalink2 import to_alphalink2
from .to_proxl import to_proxl



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

from ..data import check_input
from ..data import check_input_multi
from ..data import create_parser_result
from .filter import filter_target_decoy

from typing import Dict
from typing import List
from typing import Any


def targets_only(
    data: List[Dict[str, Any]] | Dict[str, Any],
) -> List[Dict[str, Any]] | Dict[str, Any]:
    r"""Get target crosslinks or crosslink-spectrum-matches.

    Get target crosslinks or crosslink-spectrum-matches from a list of target and decoy crosslinks or
    crosslink-spectrum-matches, or a parser_result. This effectively filters out any target-decoy and
    decoy-decoy matches and is essentially a convenience wrapper for
    ``transform.filter_target_decoy()["Target-Target"]``.

    Parameters
    ----------
    data : dict of str, any, or list of dict of str, any
        A list of crosslink-spectrum-matches or crosslinks, or a parser_result.

    Returns
    -------
    list of dict of str, any, or dict of str, any
        If a list of crosslink-spectrum-matches or crosslinks was provided, a list of target
        crosslink-spectrum-matches or crosslinks is returned. If a parser_result was provided,
        a parser_result with target crosslink-spectrum-matches and/or target crosslinks will
        be returned.

    Raises
    ------
    TypeError
        If a wrong data type is provided.
    RuntimeError
        If no target crosslinks or crosslink-spectrum-matches were found.

    Examples
    --------
    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import targets_only
    >>> result = read(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> targets = targets_only(result["crosslink-spectrum-matches"])
    >>> len(targets)
    786

    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import targets_only
    >>> result = read(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> targets = targets_only(result["crosslinks"])
    >>> len(targets)
    265

    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import targets_only
    >>> result = read(
    ...     [
    ...         "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
    ...         "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
    ...     ],
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> result_targets = targets_only(result)
    >>> len(result_targets["crosslink-spectrum-matches"])
    786
    >>> len(result_targets["crosslinks"])
    265
    """
    _ok = check_input_multi(data, "data", [dict, list])
    if isinstance(data, list):
        _ok = check_input(data, "data", list, dict)
        if len(data) == 0:
            return data
        if "data_type" not in data[0] or data[0]["data_type"] not in [
            "crosslink",
            "crosslink-spectrum-match",
        ]:
            raise TypeError(
                "Unsupported data type for input data! Parameter data has to be a list of crosslink or crosslink-spectrum-match, "
                "or a parser_result!"
            )
        targets = filter_target_decoy(data)["Target-Target"]
        if len(targets) == 0:
            raise RuntimeError(
                "No target matches found! Are you sure your data is labelled?"
            )
        return targets
    if "data_type" not in data or data["data_type"] != "parser_result":
        raise TypeError(
            "Can't filter target matches for dict. Dict has to be a valid 'parser_result'!"
        )
    new_csms = (
        filter_target_decoy(data["crosslink-spectrum-matches"])["Target-Target"]
        if data["crosslink-spectrum-matches"] is not None
        else None
    )
    new_xls = (
        filter_target_decoy(data["crosslinks"])["Target-Target"]
        if data["crosslinks"] is not None
        else None
    )
    if new_csms is not None:
        if not isinstance(new_csms, list):
            raise RuntimeError(
                "Something went wrong while getting target crosslink-spectrum-matches.\n"
                f"Expected data type: list. Got: {type(new_csms)}."
            )
        if len(new_csms) == 0:
            raise RuntimeError(
                "No target crosslink-spectrum-matches found! Are you sure they are labelled?"
            )
    if new_xls is not None:
        if not isinstance(new_xls, list):
            raise RuntimeError(
                "Something went wrong while getting target crosslinks.\n"
                f"Expected data type: list. Got: {type(new_xls)}."
            )
        if len(new_xls) == 0:
            raise RuntimeError(
                "No target crosslinks found! Are you sure they are labelled?"
            )
    return create_parser_result(
        search_engine=data["search_engine"], csms=new_csms, crosslinks=new_xls
    )



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import pandas as pd
from tqdm import tqdm

from ..data import check_input
from ..data import create_csm
from ..data import create_parser_result
from ..constants import MODIFICATIONS
from .util import format_sequence
from .util import __serialize_pandas_series

from typing import Optional
from typing import BinaryIO
from typing import Dict
from typing import Any
from typing import Tuple
from typing import List


def parse_modifications_from_maxquant_sequence(
    seq: str,
    crosslink_position: int,
    crosslinker: str,
    crosslinker_mass: float,
    modifications: Dict[str, float] = MODIFICATIONS,
) -> Dict[int, Tuple[str, float]]:
    r"""Parse post-translational-modifications from a MaxQuant peptide sequence.

    Parses post-translational-modifications (PTMs) from a MaxQuant peptide sequence,
    for example "_VVDELVKVM(Oxidation (M))GR_".

    Parameters
    ----------
    seq : str
        The MaxQuant sequence string.
    crosslink_position : int
        Position of the crosslinker in the sequence (1-based).
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    crosslinker_mass : float
        Monoisotopic delta mass of the crosslink modification.
    modifications: dict of str, float, default = ``constants.MODIFICATIONS``
        Mapping of modification names to modification masses.

    Returns
    -------
    dict of int, tuple
        The ``pyXLMS`` specific modifications object, a dictionary that maps positions to their corresponding modifications and their
        monoisotopic masses.

    Raises
    ------
    RuntimeError
        If the sequence could not be parsed because it is not in MaxQuant format.
    RuntimeError
        If multiple modifications on the same residue are parsed.
    KeyError
        If an unknown modification is encountered.

    Examples
    --------
    >>> from pyXLMS.parser import parse_modifications_from_maxquant_sequence
    >>> seq = "_VVDELVKVM(Oxidation (M))GR_"
    >>> parse_modifications_from_maxquant_sequence(seq, 2, "DSS", 138.06808)
    {2: ('DSS', 138.06808), 9: ('Oxidation', 15.994915)}

    >>> from pyXLMS.parser import parse_modifications_from_maxquant_sequence
    >>> seq = "_VVDELVKVM(Oxidation (M))GRM(Oxidation (M))_"
    >>> parse_modifications_from_maxquant_sequence(seq, 2, "DSS", 138.06808)
    {2: ('DSS', 138.06808), 9: ('Oxidation', 15.994915), 12: ('Oxidation', 15.994915)}

    >>> from pyXLMS.parser import parse_modifications_from_maxquant_sequence
    >>> seq = "_M(Oxidation (M))VVDELVKVM(Oxidation (M))GRM(Oxidation (M))_"
    >>> parse_modifications_from_maxquant_sequence(seq, 2, "DSS", 138.06808)
    {2: ('DSS', 138.06808), 1: ('Oxidation', 15.994915), 10: ('Oxidation', 15.994915), 13: ('Oxidation', 15.994915)}
    """
    parsed_modifications = {crosslink_position: (crosslinker, crosslinker_mass)}
    ## start parse seq
    split_seq = seq.split("_")
    if len(split_seq) != 3:
        raise RuntimeError(
            f"Could not parse sequence {seq}. Is the sequence correctly formatted?"
        )
    _n_term = split_seq[
        0
    ].strip()  # don't use nterm mods because I don't know how they are formatted
    internal = split_seq[1].strip()
    _c_term = split_seq[
        2
    ].strip()  # don't use cterm mods because I don't know how they are formatted
    ## end parse seq
    is_mod = 0
    current_pos = 0
    current_mod = ""
    for aa in internal:
        if is_mod == 0:
            if aa == "(":
                is_mod += 1
            else:
                current_pos += 1
        else:
            if aa == "(":
                is_mod += 1
            elif aa == ")":
                is_mod -= 1
            else:
                current_mod += aa
            if is_mod == 0:
                if current_pos in parsed_modifications:
                    raise RuntimeError(
                        f"Modification at position {current_pos} already exists!"
                    )
                else:
                    current_mod = current_mod.split()[0]
                    if current_mod not in modifications:
                        raise KeyError(
                            f"Key {current_mod} not found in parameter 'modifications'. Are you missing a modification?"
                        )
                    else:
                        parsed_modifications[current_pos] = (
                            current_mod,
                            modifications[current_mod],
                        )
                current_mod = ""
    return parsed_modifications


def read_maxquant(
    files: str | List[str] | BinaryIO,
    crosslinker: str,
    crosslinker_mass: Optional[float] = None,
    decoy_prefix: str = "REV__",
    parse_modifications: bool = True,
    modifications: Dict[str, float] = MODIFICATIONS,
    sep: str = "\t",
    decimal: str = ".",
) -> Dict[str, Any]:
    r"""Read a MaxQuant result file.

    Reads a MaxQuant crosslink-spectrum-matches result file "crosslinkMsms.txt" in ``.txt`` (tab delimited) format
    and returns a ``parser_result``.

    Parameters
    ----------
    files : str, list of str, or file stream
        The name/path of the MaxQuant result file(s) or a file-like object/stream.
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    crosslinker_mass : float, or None, default = None
        Monoisotopic delta mass of the crosslink modification. If the crosslinker is
        defined in parameter "modifications" this can be omitted.
    decoy_prefix : str, default = "REV\_\_"
        The prefix that indicates that a protein is from the decoy database.
    parse_modifications : bool, default = True
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modifications' parameter.
    modifications: dict of str, float, default = ``constants.MODIFICATIONS``
        Mapping of modification names to modification masses.
    sep : str, default = "\t"
        Seperator used in the ``.txt`` file.
    decimal : str, default = "."
        Character to recognize as decimal point.

    Returns
    -------
    dict
        The ``parser_result`` object containing all parsed information.

    Raises
    ------
    RuntimeError
        If the file(s) could not be read or if the file(s) contain no crosslink-spectrum-matches.
    KeyError
        If the specified crosslinker could not be found/mapped.

    Notes
    -----
    Uses ``Partial score 1`` as the score for the alpha peptide, ``Partial score 2`` as the score of the
    beta peptide, and ``Score`` as the score of the crosslink-spectrum-match.

    Warnings
    --------
    MaxLynx/MaxQuant only reports a single protein crosslink position per peptide, for ambiguous peptides
    only the crosslink position of the first matching protein is reported. All matching proteins can be
    retrieved via ``additional_information``, however not their corresponding crosslink positions. For this
    reason it is recommended to use ``transform.reannotate_positions()`` to correctly annotate all crosslink
    positions for all peptides if that is important for downstream analysis.

    Examples
    --------
    >>> from pyXLMS.parser import read_maxquant
    >>> csms = read_maxquant("data/maxquant/run1/crosslinkMsms.txt")
    """
    ## check input
    _ok = check_input(crosslinker, "crosslinker", str)
    _ok = (
        check_input(crosslinker_mass, "crosslinker_mass", float)
        if crosslinker_mass is not None
        else True
    )
    _ok = check_input(decoy_prefix, "decoy_prefix", str)
    _ok = check_input(parse_modifications, "parse_modifications", bool)
    _ok = check_input(modifications, "modifications", dict, float)
    _ok = check_input(sep, "sep", str)
    _ok = check_input(decimal, "decimal", str)
    if crosslinker_mass is None:
        if crosslinker not in modifications:
            if parse_modifications:
                raise KeyError(
                    "Cannot infer crosslinker mass because crosslinker is not defined in "
                    "parameter 'modifications'. Please specify crosslinker mass manually!"
                )
            else:
                crosslinker_mass = 0.0
        else:
            crosslinker_mass = modifications[crosslinker]

    ## data structures
    csms = list()

    ## handle input
    if not isinstance(files, list):
        inputs = [files]
    else:
        inputs = files

    ## process data
    for input in inputs:
        data = pd.read_csv(input, sep=sep, decimal=decimal, low_memory=False)
        xl = data.dropna(axis=0, subset=["Proteins2"])
        for i, row in tqdm(
            xl.iterrows(), total=xl.shape[0], desc="Reading MaxQuant CSMs..."
        ):
            # preprocess proteins
            protein_a = (
                str(row["Proteins1"]).split("(")[0].strip()
                if "(" in str(row["Proteins1"])
                else str(row["Proteins1"])
            )
            protein_b = (
                str(row["Proteins2"]).split("(")[0].strip()
                if "(" in str(row["Proteins2"])
                else str(row["Proteins2"])
            )
            # create csm
            csm = create_csm(
                peptide_a=format_sequence(str(row["Sequence1"])),
                modifications_a=parse_modifications_from_maxquant_sequence(
                    str(row["Modified sequence1"]),
                    int(row["Peptide index of Crosslink 1"]),
                    crosslinker,
                    crosslinker_mass,
                    modifications,
                )
                if parse_modifications
                else None,
                xl_position_peptide_a=int(row["Peptide index of Crosslink 1"]),
                proteins_a=[
                    protein_a.strip()
                    if protein_a.strip()[: len(decoy_prefix)] != decoy_prefix
                    else protein_a.strip()[len(decoy_prefix) :]
                ],
                xl_position_proteins_a=[int(row["Protein index of Crosslink 1"])],
                pep_position_proteins_a=[
                    int(row["Protein index of Crosslink 1"])
                    - int(row["Peptide index of Crosslink 1"])
                    + 1
                ],
                score_a=float(row["Partial score 1"]),
                decoy_a=decoy_prefix in str(row["Proteins1"]),
                peptide_b=format_sequence(str(row["Sequence2"])),
                modifications_b=parse_modifications_from_maxquant_sequence(
                    str(row["Modified sequence2"]),
                    int(row["Peptide index of Crosslink 2"]),
                    crosslinker,
                    crosslinker_mass,
                    modifications,
                )
                if parse_modifications
                else None,
                xl_position_peptide_b=int(row["Peptide index of Crosslink 2"]),
                proteins_b=[
                    protein_b.strip()
                    if protein_b.strip()[: len(decoy_prefix)] != decoy_prefix
                    else protein_b.strip()[len(decoy_prefix) :]
                ],
                xl_position_proteins_b=[int(row["Protein index of Crosslink 2"])],
                pep_position_proteins_b=[
                    int(row["Protein index of Crosslink 2"])
                    - int(row["Peptide index of Crosslink 2"])
                    + 1
                ],
                score_b=float(row["Partial score 2"]),
                decoy_b=decoy_prefix in str(row["Proteins2"]),
                score=float(row["Score"]),
                spectrum_file=str(row["Raw file"]).strip(),
                scan_nr=int(row["Scan number"]),
                charge=int(row["Charge"]),
                rt=None,
                im_cv=None,
                additional_information={
                    "source": __serialize_pandas_series(row),
                    "Proteins1": str(row["Proteins1"]).strip(),
                    "Proteins2": str(row["Proteins2"]).strip(),
                    "Delta score": float(row["Delta score"]),
                },
            )
            csms.append(csm)
    ## check results
    if len(csms) == 0:
        raise RuntimeError(
            "No crosslink-spectrum-matches were parsed! If this is unexpected, please file a bug report!"
        )
    ## return parser result
    return create_parser_result(
        search_engine="MaxQuant",
        csms=csms,
        crosslinks=None,
    )


def read_maxlynx(
    files: str | List[str] | BinaryIO,
    crosslinker: str,
    crosslinker_mass: Optional[float] = None,
    decoy_prefix: str = "REV__",
    parse_modifications: bool = True,
    modifications: Dict[str, float] = MODIFICATIONS,
    sep: str = "\t",
    decimal: str = ".",
) -> Dict[str, Any]:
    r"""Read a MaxLynx result file.

    Reads a MaxLynx crosslink-spectrum-matches result file "crosslinkMsms.txt" in ``.txt`` (tab delimited) format
    and returns a ``parser_result``. This is an alias for the MaxQuant reader.

    Parameters
    ----------
    files : str, list of str, or file stream
        The name/path of the MaxLynx result file(s) or a file-like object/stream.
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    crosslinker_mass : float, or None, default = None
        Monoisotopic delta mass of the crosslink modification. If the crosslinker is
        defined in parameter "modifications" this can be omitted.
    decoy_prefix : str, default = "REV\_\_"
        The prefix that indicates that a protein is from the decoy database.
    parse_modifications : bool, default = True
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modifications' parameter.
    modifications: dict of str, float, default = ``constants.MODIFICATIONS``
        Mapping of modification names to modification masses.
    sep : str, default = "\t"
        Seperator used in the ``.txt`` file.
    decimal : str, default = "."
        Character to recognize as decimal point.

    Returns
    -------
    dict
        The ``parser_result`` object containing all parsed information.

    Raises
    ------
    RuntimeError
        If the file(s) could not be read or if the file(s) contain no crosslink-spectrum-matches.
    KeyError
        If the specified crosslinker could not be found/mapped.

    Notes
    -----
    Uses ``Partial score 1`` as the score for the alpha peptide, ``Partial score 2`` as the score of the
    beta peptide, and ``Score`` as the score of the crosslink-spectrum-match.

    Warnings
    --------
    MaxLynx/MaxQuant only reports a single protein crosslink position per peptide, for ambiguous peptides
    only the crosslink position of the first matching protein is reported. All matching proteins can be
    retrieved via ``additional_information``, however not their corresponding crosslink positions. For this
    reason it is recommended to use ``transform.reannotate_positions()`` to correctly annotate all crosslink
    positions for all peptides if that is important for downstream analysis.

    Examples
    --------
    >>> from pyXLMS.parser import read_maxlynx
    >>> csms_from_xlsx = read_maxlynx("data/maxquant/run1/crosslinkMsms.txt")
    """
    return read_maxquant(
        files,
        crosslinker,
        crosslinker_mass,
        decoy_prefix,
        parse_modifications,
        modifications,
        sep,
        decimal,
    )



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest
import pandas as pd


YHU_PDB = "data/_test/exporter/pyxlinkviewer/6YHU/6yhu.pdb"
YHU_RESULT_ALL = (
    "data/_test/exporter/pyxlinkviewer/6YHU/all/unique_links_all_crosslinks.txt"
)


def helper_get_dataframe_from_pyxlinkviewer(filename: str) -> pd.DataFrame:
    pyxlinkviewer = ""
    with open(filename, "r", encoding="utf-8") as f:
        pyxlinkviewer = f.read()
        f.close()
    pos_a = list()
    chain_a = list()
    pos_b = list()
    chain_b = list()
    for line in pyxlinkviewer.split("\n"):
        if len(line.strip()) > 0:
            xl = line.split("|")
            pos_a.append(int(xl[0]))
            chain_a.append(xl[1].strip())
            pos_b.append(int(xl[2]))
            chain_b.append(xl[3].strip())
    return pd.DataFrame(
        {"residue 1": pos_a, "chain 1": chain_a, "residue 2": pos_b, "chain 2": chain_b}
    )


def test1():
    from pyXLMS.exporter import to_xlmstools
    from pyXLMS.parser import read_custom

    pr = read_custom("data/_test/exporter/xlms-tools/unique_links_all_pyxlms.csv")
    crosslinks = pr["crosslinks"]
    xlmstools_result = to_xlmstools(crosslinks, pdb_file="6YHU", filename_prefix="6YHU")
    xlmstools_output_file_str = xlmstools_result["xlms-tools"]
    xlmstools_dataframe = xlmstools_result["xlms-tools DataFrame"]
    nr_mapped_crosslinks = xlmstools_result["Number of mapped crosslinks"]
    crosslink_mapping = xlmstools_result["Mapping"]
    parsed_pdb_sequenece = xlmstools_result["Parsed PDB sequence"]
    parsed_pdb_chains = xlmstools_result["Parsed PDB chains"]
    parsed_pdb_residue_numbers = xlmstools_result["Parsed PDB residue numbers"]
    exported_files = xlmstools_result["Exported files"]
    assert exported_files[0] == "6YHU_xlms-tools.txt"
    assert xlmstools_output_file_str is not None
    assert xlmstools_dataframe is not None
    assert nr_mapped_crosslinks is not None
    assert crosslink_mapping is not None
    assert parsed_pdb_sequenece is not None
    assert parsed_pdb_chains is not None
    assert parsed_pdb_residue_numbers is not None


def test2():
    import os
    from pyXLMS.exporter import to_xlmstools
    from pyXLMS.parser import read_custom

    pr = read_custom("data/_test/exporter/xlms-tools/unique_links_all_pyxlms.csv")
    crosslinks = pr["crosslinks"]
    xlmstools_result = to_xlmstools(crosslinks, pdb_file="6YHU", filename_prefix="test")
    assert xlmstools_result is not None
    xlmstools_output_file_str = xlmstools_result["xlms-tools"]
    assert xlmstools_output_file_str is not None
    xlmstools_dataframe = xlmstools_result["xlms-tools DataFrame"]
    assert xlmstools_dataframe is not None
    nr_mapped_crosslinks = xlmstools_result["Number of mapped crosslinks"]
    assert nr_mapped_crosslinks == 28
    crosslink_mapping = xlmstools_result["Mapping"]
    assert crosslink_mapping is not None
    parsed_pdb_sequenece = xlmstools_result["Parsed PDB sequence"]
    assert parsed_pdb_sequenece is not None
    parsed_pdb_chains = xlmstools_result["Parsed PDB chains"]
    assert parsed_pdb_chains is not None
    parsed_pdb_residue_numbers = xlmstools_result["Parsed PDB residue numbers"]
    assert parsed_pdb_residue_numbers is not None
    exported_files = xlmstools_result["Exported files"]
    assert len(exported_files) == 4
    assert exported_files[0] == "test_xlms-tools.txt"
    for exported_file in exported_files:
        assert os.path.isfile(exported_file)


def test3():
    from pyXLMS.exporter import to_xlmstools
    from pyXLMS.parser import read_custom

    pr = read_custom("data/_test/exporter/xlms-tools/unique_links_all_pyxlms.csv")
    crosslinks = pr["crosslinks"]
    xlmstools_result = to_xlmstools(crosslinks, pdb_file="6YHU")
    should_result = helper_get_dataframe_from_pyxlinkviewer(YHU_RESULT_ALL)
    assert should_result.equals(xlmstools_result["xlms-tools DataFrame"])


def test4():
    from pyXLMS.exporter import to_xlmstools
    from pyXLMS.parser import read_custom

    pr = read_custom("data/_test/exporter/xlms-tools/unique_links_all_pyxlms.csv")
    crosslinks = pr["crosslinks"]
    xlmstools_result = to_xlmstools(crosslinks, pdb_file=YHU_PDB)
    should_result = helper_get_dataframe_from_pyxlinkviewer(YHU_RESULT_ALL)
    assert should_result.equals(xlmstools_result["xlms-tools DataFrame"])


def test5():
    from pyXLMS.exporter import to_xlmstools
    from pyXLMS.parser import read_custom

    pr = read_custom("data/_test/exporter/xlms-tools/unique_links_all_pyxlms.csv")
    crosslinks = pr["crosslinks"]
    with pytest.raises(
        ValueError,
        match=r"Minimum sequence identity should be given as a fraction, e\.g\. 0\.8 for 80\% minimum sequence identity!",
    ):
        _xlmstools_result = to_xlmstools(
            crosslinks, pdb_file="6YHU", min_sequence_identity=80.0
        )


def test6():
    from pyXLMS.exporter import to_xlmstools

    with pytest.raises(ValueError, match=r"Provided crosslinks contain no elements!"):
        _xlmstools_result = to_xlmstools([], pdb_file="6YHU")


def test7():
    from pyXLMS.exporter import to_xlmstools
    from pyXLMS.data import create_csm_min

    csm1 = create_csm_min("KPEPTIDE", 1, "PKEPTIDE", 2, "RUN_1", 1)
    csms = [csm1]
    with pytest.raises(
        TypeError,
        match="Unsupported data type for input crosslinks! Parameter crosslinks has to be a list of crosslinks!",
    ):
        _xlmstools_result = to_xlmstools(csms, pdb_file="6YHU")


def test14():
    from pyXLMS.exporter import to_xlmstools
    from pyXLMS.data import create_csm_min
    from pyXLMS.data import create_crosslink_min

    csm1 = create_csm_min("KPEPTIDE", 1, "PKEPTIDE", 2, "RUN_1", 1)
    xl1 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    mixed = [xl1, csm1]
    with pytest.raises(
        TypeError, match="Not all elements in data have the same data type!"
    ):
        _xlmstools_result = to_xlmstools(mixed, pdb_file="6YHU")



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import sqlite3
import warnings
import pandas as pd
from tqdm import tqdm
from os.path import splitext

from ..data import check_input
from ..data import create_crosslink
from ..data import create_csm
from ..data import create_parser_result
from ..constants import MODIFICATIONS
from .util import format_sequence
from .util import get_bool_from_value
from .util import __serialize_pandas_series

from typing import BinaryIO
from typing import Dict
from typing import Any
from typing import Tuple
from typing import List

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def __check_positions_okay(positions: List[int]) -> bool:
    r"""Checks if all the positions are positive.

    Parameters
    ----------
    positions : list of int
        List of positions.

    Returns
    -------
    bool
        If all positions are valid (greater than zero).

    Notes
    -----
    This function should not be called directly, it is called from ``read_msannika()``.
    """
    for position in positions:
        if position < 1:
            return False
    return True


def __read_msannika_pdresult(filename: str, drop: bool = False) -> List[pd.DataFrame]:
    r"""Read an MS Annika pdResult file and convert it to standard MS Annika result tables.

    Reads an MS Annika pdResult file and converts it to standard MS Annika CSM and crosslink
    tables as pandas DataFrame objects.

    Parameters
    ----------
    filename : str
        Filename/path of the pdResult file.
    drop : bool, default = False
        Whether or not the not-needed columns should be dropped.
        Defaults to not dropping not-needed columns.

    Returns
    -------
    list of pd.DataFrame
        The converted CSM and crosslink tables in MS Annika format.

    Notes
    -----
    This function should not be called directly, it is called from ``read_msannika()``.
    """
    conn = sqlite3.connect(filename)
    csms = pd.read_sql_query("SELECT * FROM CSMs", conn)
    xls = pd.read_sql_query("SELECT * FROM Crosslinks", conn)
    conn.close()
    column_mapping_csms = {
        "SequenceA": "Sequence A",
        "SequenceB": "Sequence B",
        "ModificationsA": "Modifications A",
        "ModificationsB": "Modifications B",
        "CrosslinkerPositionA": "Crosslinker Position A",
        "CrosslinkerPositionB": "Crosslinker Position B",
        "AccessionA": "Accession A",
        "AccessionB": "Accession B",
        "Ainprotein": "A in protein",
        "Binprotein": "B in protein",
        "ScoreAlpha": "Score Alpha",
        "ScoreBeta": "Score Beta",
        "AlphaTD": "Alpha T/D",
        "BetaTD": "Beta T/D",
        "CombinedScore": "Combined Score",
        "SpectrumFileName": "Spectrum File",
        "FirstScan": "First Scan",
        "ChargeState": "Charge",
        "RetentionTime": "RT [min]",
        "CompensationVoltage": "Compensation Voltage",
    }
    column_mapping_xls = {
        "SequenceA": "Sequence A",
        "SequenceB": "Sequence B",
        "PositionA": "Position A",
        "PositionB": "Position B",
        "AccessionA": "Accession A",
        "AccessionB": "Accession B",
        "InproteinA": "In protein A",
        "InproteinB": "In protein B",
        "Decoy": "Decoy",
        "BestCSMScore": "Best CSM Score",
    }
    csms.rename(columns=column_mapping_csms, inplace=True)
    xls.rename(columns=column_mapping_xls, inplace=True)
    if drop:
        csms.drop(
            columns=list(
                set(csms.columns.values.tolist())
                - set(list(column_mapping_csms.values()))
            ),
            inplace=True,
        )
        xls.drop(
            columns=list(
                set(xls.columns.values.tolist())
                - set(list(column_mapping_xls.values()))
            ),
            inplace=True,
        )
    return [csms, xls]


def read_msannika(
    files: str | List[str] | BinaryIO,
    parse_modifications: bool = True,
    modifications: Dict[str, float] = MODIFICATIONS,
    format: Literal["auto", "csv", "txt", "tsv", "xlsx", "pdresult"] = "auto",
    sep: str = "\t",
    decimal: str = ".",
    unsafe: bool = False,
    verbose: Literal[0, 1, 2] = 1,
) -> Dict[str, Any]:
    r"""Read an MS Annika result file.

    Reads an MS Annika crosslink-spectrum-matches result file or crosslink result file in ``.csv`` or ``.xlsx`` format,
    or both from a ``.pdResult`` file from Proteome Discover, and returns a ``parser_result``.

    Parameters
    ----------
    files : str, list of str, or file stream
        The name/path of the MS Annika result file(s) or a file-like object/stream.
    parse_modifications : bool, default = True
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modifications' parameter.
    modifications: dict of str, float, default = ``constants.MODIFICATIONS``
        Mapping of modification names to modification masses.
    format : "auto", "csv", "tsv", "txt", "xlsx", or "pdresult", default = "auto"
        The format of the result file. ``"auto"`` is only available if the name/path to the MS Annika result file is given.
    sep : str, default = "\t"
        Seperator used in the ``.csv`` or ``.tsv`` file. Parameter is ignored if the file is in ``.xlsx`` or ``.pdResult`` format.
    decimal : str, default = "."
        Character to recognize as decimal point. Parameter is ignored if the file is in ``.xlsx`` or ``.pdResult`` format.
    unsafe : bool, default = False
        If True, allows reading of negative peptide and crosslink positions but replaces their values with None.
        Negative values occur when peptides can't be matched to proteins because of 'X' in protein sequences.
        Reannotation might be possible with ``transform.reannotate_positions()``.
    verbose : 0, 1, or 2, default = 1
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    dict
        The ``parser_result`` object containing all parsed information.

    Raises
    ------
    ValueError
        If the input format is not supported or cannot be inferred.
    TypeError
        If the pdResult file is provided in the wrong format.
    TypeError
        If parameter verbose was not set correctly.
    RuntimeError
        If one of the crosslinks or crosslink-spectrum-matches contains unknown crosslink or peptide positions.
        This occurs when peptides can't be matched to proteins because of 'X' in protein sequences. Selecting
        'unsafe = True' will ignore these errors and return None type positions.
        Reannotation might be possible with ``transform.reannotate_positions()``.
    RuntimeError
        If the file(s) could not be read or if the file(s) contain no crosslinks or crosslink-spectrum-matches.
    KeyError
        If one of the found post-translational-modifications could not be found/mapped.

    Notes
    -----
    Uses ``Score Alpha`` as the score for the alpha peptide, ``Score Beta`` as the score of the
    beta peptide, ``Combined Score`` as the score of the crosslink-spectrum-match, and ``Best CSM Score``
    as the score of the crosslink.

    Warnings
    --------
    MS Annika does not report if the individual peptides in a crosslink are from the target or decoy database.
    The parser assumes that both peptides from a target crosslink are from the target database, and vice versa,
    that both peptides are from the decoy database if it is a decoy crosslink. This leads to only TT and DD matches,
    which needs to be considered for FDR estimation. This also only applies to crosslinks and not crosslink-spectrum-matches,
    where this information is correctly reported and parsed.

    Examples
    --------
    >>> from pyXLMS.parser import read_msannika
    >>> csms_from_xlsx = read_msannika(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx"
    ... )

    >>> from pyXLMS.parser import read_msannika
    >>> crosslinks_from_xlsx = read_msannika(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx"
    ... )

    >>> from pyXLMS.parser import read_msannika
    >>> csms_from_tsv = read_msannika(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt"
    ... )

    >>> from pyXLMS.parser import read_msannika
    >>> crosslinks_from_tsv = read_msannika(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.txt"
    ... )

    >>> from pyXLMS.parser import read_msannika
    >>> csms_and_crosslinks_from_pdresult = read_msannika(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult"
    ... )
    """
    ## check input
    _ok = check_input(parse_modifications, "parse_modifications", bool)
    _ok = check_input(modifications, "modifications", dict, float)
    _ok = check_input(format, "format", str)
    _ok = check_input(sep, "sep", str)
    _ok = check_input(decimal, "decimal", str)
    _ok = check_input(unsafe, "unsafe", bool)
    _ok = check_input(verbose, "verbose", int)
    if verbose not in [0, 1, 2]:
        raise TypeError("Verbose level has to be one of 0, 1, or 2!")

    ## helper functions
    def parse_modification_str(
        sequence: str,
        modification_str: str,
        modifications: Dict[str, float] = modifications,
    ) -> Dict[int, Tuple[str, float]]:
        mods = [mod.strip() for mod in modification_str.split(";")]
        parsed_mods = dict()
        for mod in mods:
            mod_type = mod.split("(")[1].split(")")[0].strip()
            mod_pos = mod.split("(")[0].strip()
            if mod_type not in modifications:
                raise KeyError(
                    f"Unable to find modification {mod_type} in the set of provided modifications. "
                    + "Please pass the full set of expected modifications to the parser."
                )
            if "Nterm" in mod_pos or "N-Term" in mod_pos:
                parsed_mods[0] = (mod_type, modifications[mod_type])
            elif "Cterm" in mod_pos or "C-Term" in mod_pos:
                parsed_mods[len(sequence)] = (mod_type, modifications[mod_type])
            else:
                parsed_mods[int(mod_pos[1:])] = (mod_type, modifications[mod_type])
        return parsed_mods

    ## data structures
    crosslinks = list()
    csms = list()

    ## handle input
    if not isinstance(files, list):
        inputs = [files]
    else:
        inputs = files

    for input in inputs:
        ## reading data
        data_objects = None
        if format == "auto" and not isinstance(input, str):
            raise ValueError(
                "Can't detect format for file-like objects. Please specify format manually!"
            )
        # and isinstance specified for type checking
        if format == "auto" and isinstance(input, str):
            file_extension = splitext(input)[1].lower()
            if (
                file_extension == ".txt"
                or file_extension == ".tsv"
                or file_extension == ".csv"
            ):
                data_objects = [
                    pd.read_csv(input, sep=sep, decimal=decimal, low_memory=False)
                ]
            elif file_extension == ".xlsx":
                data_objects = [pd.read_excel(input, engine="openpyxl")]
            elif file_extension == ".pdresult":
                data_objects = __read_msannika_pdresult(input)
            else:
                raise ValueError(
                    f"Detected file extension {file_extension} is not supported! Input file has to be a valid file with extension '.csv', '.tsv', '.xlsx', or '.pdResult'!"
                )
        elif format in ["csv", "tsv", "txt", "xlsx"]:
            if format == "xlsx":
                data_objects = [pd.read_excel(input, engine="openpyxl")]
            else:
                data_objects = [
                    pd.read_csv(input, sep=sep, decimal=decimal, low_memory=False)
                ]
        elif format == "pdresult":
            if not isinstance(input, str):
                raise TypeError(
                    "Can't read pdResult files from a file-like object/stream. Please provide the filename/path instead!"
                )
            data_objects = __read_msannika_pdresult(input)
        else:
            raise ValueError(
                f"Provided input format {format} is not supported! Input format has to be of type 'csv', 'tsv', 'xlsx', or 'pdresult'!"
            )
        if data_objects is None:
            raise RuntimeError(
                "Something went wrong while reading the file! Please file a bug report!"
            )
        # this should be impossible, but check here for pyright
        if not isinstance(data_objects, list):
            raise RuntimeError(
                "Something went wrong while reading the file! Please file a bug report!"
            )
        for data in data_objects:
            # this should be impossible, but check here for pyright
            if not isinstance(data, pd.DataFrame):
                raise RuntimeError(
                    "Something went wrong while reading the file! Please file a bug report!"
                )
            ## detect input file type
            col_names = data.columns.values.tolist()
            is_crosslink_dataframe = "Best CSM Score" in col_names
            ## process data
            if is_crosslink_dataframe:
                for i, row in tqdm(
                    data.iterrows(),
                    total=data.shape[0],
                    desc="Reading MS Annika crosslinks...",
                ):
                    # pre compute values
                    xl_position_proteins_a = [
                        int(position)
                        for position in str(row["In protein A"]).split(";")
                    ]
                    if not __check_positions_okay(xl_position_proteins_a):
                        if unsafe and not verbose == 2:
                            xl_position_proteins_a = None
                            if verbose == 1:
                                warnings.warn(
                                    RuntimeWarning(
                                        f"Encountered invalid crosslink position for crosslink with sequence "
                                        f"{format_sequence(str(row['Sequence A']))}-{format_sequence(str(row['Sequence B']))}!"
                                    )
                                )
                        else:
                            raise RuntimeError(
                                f"Encountered invalid crosslink position for crosslink with sequence "
                                f"{format_sequence(str(row['Sequence A']))}-{format_sequence(str(row['Sequence B']))}!"
                            )
                    xl_position_proteins_b = [
                        int(position)
                        for position in str(row["In protein B"]).split(";")
                    ]
                    if not __check_positions_okay(xl_position_proteins_b):
                        if unsafe and not verbose == 2:
                            xl_position_proteins_b = None
                            if verbose == 1:
                                warnings.warn(
                                    RuntimeWarning(
                                        f"Encountered invalid crosslink position for crosslink with sequence "
                                        f"{format_sequence(str(row['Sequence A']))}-{format_sequence(str(row['Sequence B']))}!"
                                    )
                                )
                        else:
                            raise RuntimeError(
                                f"Encountered invalid crosslink position for crosslink with sequence "
                                f"{format_sequence(str(row['Sequence A']))}-{format_sequence(str(row['Sequence B']))}!"
                            )
                    # create crosslink
                    crosslink = create_crosslink(
                        peptide_a=format_sequence(str(row["Sequence A"])),
                        xl_position_peptide_a=int(row["Position A"]),
                        proteins_a=[
                            protein.strip()
                            for protein in str(row["Accession A"]).split(";")
                        ],
                        xl_position_proteins_a=xl_position_proteins_a,
                        decoy_a=get_bool_from_value(row["Decoy"]),
                        peptide_b=format_sequence(str(row["Sequence B"])),
                        xl_position_peptide_b=int(row["Position B"]),
                        proteins_b=[
                            protein.strip()
                            for protein in str(row["Accession B"]).split(";")
                        ],
                        xl_position_proteins_b=xl_position_proteins_b,
                        decoy_b=get_bool_from_value(row["Decoy"]),
                        score=float(row["Best CSM Score"]),
                        additional_information={
                            "source": __serialize_pandas_series(row)
                        },
                    )
                    crosslinks.append(crosslink)
            else:
                for i, row in tqdm(
                    data.iterrows(),
                    total=data.shape[0],
                    desc="Reading MS Annika CSMs...",
                ):
                    # pre compute values
                    xl_position_proteins_a = [
                        int(position) + int(row["Crosslinker Position A"])
                        for position in str(row["A in protein"]).split(";")
                    ]
                    if not __check_positions_okay(xl_position_proteins_a):
                        if unsafe and not verbose == 2:
                            xl_position_proteins_a = None
                            if verbose == 1:
                                warnings.warn(
                                    RuntimeWarning(
                                        f"Encountered invalid crosslink position for crosslink-spectrum-match with scan number: {int(row['First Scan'])}!"
                                    )
                                )
                        else:
                            raise RuntimeError(
                                f"Encountered invalid crosslink position for crosslink-spectrum-match with scan number: {int(row['First Scan'])}!"
                            )
                    pep_position_proteins_a = [
                        int(position) + 1
                        for position in str(row["A in protein"]).split(";")
                    ]
                    if not __check_positions_okay(pep_position_proteins_a):
                        if unsafe and not verbose == 2:
                            pep_position_proteins_a = None
                            if verbose == 1:
                                warnings.warn(
                                    RuntimeWarning(
                                        f"Encountered invalid crosslink position for crosslink-spectrum-match with scan number: {int(row['First Scan'])}!"
                                    )
                                )
                        else:
                            raise RuntimeError(
                                f"Encountered invalid peptide position for crosslink-spectrum-match with scan number: {int(row['First Scan'])}!"
                            )
                    xl_position_proteins_b = [
                        int(position) + int(row["Crosslinker Position B"])
                        for position in str(row["B in protein"]).split(";")
                    ]
                    if not __check_positions_okay(xl_position_proteins_b):
                        if unsafe and not verbose == 2:
                            xl_position_proteins_b = None
                            if verbose == 1:
                                warnings.warn(
                                    RuntimeWarning(
                                        f"Encountered invalid crosslink position for crosslink-spectrum-match with scan number: {int(row['First Scan'])}!"
                                    )
                                )
                        else:
                            raise RuntimeError(
                                f"Encountered invalid crosslink position for crosslink-spectrum-match with scan number: {int(row['First Scan'])}!"
                            )
                    pep_position_proteins_b = [
                        int(position) + 1
                        for position in str(row["B in protein"]).split(";")
                    ]
                    if not __check_positions_okay(pep_position_proteins_b):
                        if unsafe and not verbose == 2:
                            pep_position_proteins_b = None
                            if verbose == 1:
                                warnings.warn(
                                    RuntimeWarning(
                                        f"Encountered invalid crosslink position for crosslink-spectrum-match with scan number: {int(row['First Scan'])}!"
                                    )
                                )
                        else:
                            raise RuntimeError(
                                f"Encountered invalid peptide position for crosslink-spectrum-match with scan number: {int(row['First Scan'])}!"
                            )
                    # create csm
                    csm = create_csm(
                        peptide_a=format_sequence(str(row["Sequence A"])),
                        modifications_a=parse_modification_str(
                            format_sequence(str(row["Sequence A"]).strip()),
                            str(row["Modifications A"]).strip(),
                        )
                        if parse_modifications
                        else None,
                        xl_position_peptide_a=int(row["Crosslinker Position A"]),
                        proteins_a=[
                            protein.strip()
                            for protein in str(row["Accession A"]).split(";")
                        ],
                        xl_position_proteins_a=xl_position_proteins_a,
                        pep_position_proteins_a=pep_position_proteins_a,
                        score_a=float(row["Score Alpha"]),
                        decoy_a=not get_bool_from_value(str(row["Alpha T/D"])),
                        peptide_b=format_sequence(str(row["Sequence B"])),
                        modifications_b=parse_modification_str(
                            format_sequence(str(row["Sequence B"]).strip()),
                            str(row["Modifications B"]).strip(),
                        )
                        if parse_modifications
                        else None,
                        xl_position_peptide_b=int(row["Crosslinker Position B"]),
                        proteins_b=[
                            protein.strip()
                            for protein in str(row["Accession B"]).split(";")
                        ],
                        xl_position_proteins_b=xl_position_proteins_b,
                        pep_position_proteins_b=pep_position_proteins_b,
                        score_b=float(row["Score Beta"]),
                        decoy_b=not get_bool_from_value(str(row["Beta T/D"])),
                        score=float(row["Combined Score"]),
                        spectrum_file=str(row["Spectrum File"]).strip(),
                        scan_nr=int(row["First Scan"]),
                        charge=int(row["Charge"]),
                        rt=float(row["RT [min]"]) * 60.0,
                        im_cv=float(row["Compensation Voltage"]),
                        additional_information={
                            "source": __serialize_pandas_series(row)
                        },
                    )
                    csms.append(csm)
    ## check results
    if len(crosslinks) + len(csms) == 0:
        raise RuntimeError(
            "No crosslink-spectrum-matches or crosslinks were parsed! If this is unexpected, please file a bug report!"
        )
    ## return parser result
    return create_parser_result(
        search_engine="MS Annika",
        csms=csms if len(csms) > 0 else None,
        crosslinks=crosslinks if len(crosslinks) > 0 else None,
    )



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

__all__ = [
    "venn",
    "plot_venn_diagram",
    "plot_score_distribution",
    "plot_target_decoy_distribution",
    "plot_protein_distribution",
    "plot_peptide_pair_distribution",
    "plot_crosslink_type_distribution",
]

from .plot_venn_diagram import venn
from .plot_venn_diagram import plot_venn_diagram
from .plot_score_distribution import plot_score_distribution
from .plot_target_decoy_distribution import plot_target_decoy_distribution
from .plot_protein_distribution import plot_protein_distribution
from .plot_peptide_pair_distribution import plot_peptide_pair_distribution
from .plot_crosslink_type_distribution import plot_crosslink_type_distribution
